# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль фильтров языковой модели по разным признакам.

from collections import defaultdict
from typing import List

import numpy as np

from rupo.main.vocabulary import StressVocabulary
from rupo.rhymes.rhymes import Rhymes
from rupo.generate.word_form_vocabulary import WordFormVocabulary
from rupo.stress.word import StressedWord


class Filter(object):
    """
    Фильтр языковой модели.
    """
    def filter_word(self, word: StressedWord) -> bool:
        raise NotImplementedError()

    def pass_word(self, word: StressedWord) -> None:
        raise NotImplementedError()

    def revert_word(self, word: StressedWord) -> None:
        raise NotImplementedError()

    def is_completed(self) -> bool:
        raise NotImplementedError()

    def filter_model(self, model: np.array, vocabulary: StressVocabulary) -> np.array:
        """
        Фильтрация языковой модели.

        :param model: изначальная модель.
        :param vocabulary: словарь
        :return: модель после фильтрации и нормирования.
        """
        for i in range(len(model)):
            if not self.filter_word(vocabulary.get_word(i)):
                model[i] = 0.0
        return model

    def filter_words(self, words: List[StressedWord]) -> List[StressedWord]:
        """
        Фильтрация набора слов.

        :param words: слова.
        :return: слова после фильтрации.
        """
        return [word for word in words if self.filter_word(word)]


class MetreFilter(Filter):
    """
    Фильтр по шаблону метра.
    """
    def __init__(self, metre_pattern: str):
        self.metre_pattern = metre_pattern
        self.position = len(metre_pattern) - 1

    def filter_word(self, word: StressedWord) -> bool:
        """
        Фильтрация слова по метру в текущей позиции.

        :param word: слово.
        :return: подходит слово или нет.
        """
        syllables = word.syllables
        syllables_count = len(syllables)
        if syllables_count == 0:
            return False
        if syllables_count > self.position + 1:
            return False
        for i in range(syllables_count):
            syllable = syllables[i]
            syllable_number = self.position - syllables_count + i + 1
            if syllables_count >= 2 and syllable.stress == -1 and self.metre_pattern[syllable_number] == "+":
                for j in range(syllables_count):
                    other_syllable = syllables[j]
                    other_syllable_number = other_syllable.number - syllable.number + syllable_number
                    if i != j and other_syllable.stress != -1 and self.metre_pattern[other_syllable_number] == "-":
                        return False
        return True

    def pass_word(self, word: StressedWord) -> None:
        """
        Сдвинуть позицию в шаблоне метра на слово.

        :param word: слово.
        """
        self.position -= len(word.syllables)

    def revert_word(self, word: StressedWord) -> None:
        """
        Сдвинуть позицию в шаблоне метра на слово назад.

        :param word: слово.
        """
        self.position += len(word.syllables)

    def reset(self) -> None:
        """
        Сброс позиции в шаблоне.
        """
        self.position = len(self.metre_pattern) - 1

    def is_completed(self):
        """
        :return: закончена ли генерация по фильтру?
        """
        return self.position < 0


class RhymeFilter(Filter):
    """
    Фильтр по шаблону рифмы.
    """
    def __init__(self, rhyme_pattern: str, letters_to_rhymes: dict=None,
                 word_form_vocabulary: WordFormVocabulary=None, score_border=4):
        self.word_form_vocabulary = word_form_vocabulary  # type: WordFormVocabulary
        self.rhyme_pattern = rhyme_pattern
        self.position = len(self.rhyme_pattern) - 1
        self.letters_to_rhymes = defaultdict(set)
        self.score_border = score_border
        if letters_to_rhymes is not None:
            for letter, words in letters_to_rhymes.items():
                for word in words:
                    self.letters_to_rhymes[letter].add(word)

    def filter_word(self, word: StressedWord) -> bool:
        """
        Фильтрация слова по рифме в текущей позиции.

        :param word: слово.
        :return: подходит слово или нет.
        """
        if len(word.syllables) <= 1:
            return False
        if len(self.letters_to_rhymes[self.rhyme_pattern[self.position]]) == 0:
            return True
        first_word = list(self.letters_to_rhymes[self.rhyme_pattern[self.position]])[0]

        is_rhyme = Rhymes.is_rhyme(first_word, word, score_border=self.score_border, syllable_number_border=2,
                                   word_form_vocabulary=self.word_form_vocabulary) and \
            first_word.text != word.text
        return is_rhyme

    def pass_word(self, word: StressedWord) -> None:
        """
        Сдвинуть позицию в шаблоне рифмы на строчку.

        :param word: рифмующееся слово.
        """
        self.letters_to_rhymes[self.rhyme_pattern[self.position]].add(word)
        self.position -= 1

    def revert_word(self, word: StressedWord) -> None:
        """
        Сдвинуть позицию в шаблоне рифмы на строчку назад.

        :param word: рифмующееся слово.
        """
        self.position += 1
        self.letters_to_rhymes[self.rhyme_pattern[self.position]].remove(word)

    def is_completed(self):
        """
        :return: закончена ли генерация по фильтру?
        """
        return self.position < 0

    def reset(self) -> None:
        """
        Сброс позиции в шаблоне.
        """
        self.position = len(self.rhyme_pattern) - 1



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс рифм.

from typing import Tuple
from rupo.g2p.graphemes import Graphemes
from rupo.util.preprocess import VOWELS
from rupo.generate.word_form_vocabulary import WordFormVocabulary
from rupo.stress.word import StressedWord


class Rhymes(object):
    """
    Поиск рифм.
    """

    @staticmethod
    def is_rhyme(word1: StressedWord, word2: StressedWord, score_border: int=4, syllable_number_border: int=4,
                 word_form_vocabulary: WordFormVocabulary=None) -> bool:
        """
        Проверка рифмованности 2 слов.

        :param word1: первое слово для проверки рифмы, уже акцентуированное (Word).
        :param word2: второе слово для проверки рифмы, уже акцентуированное (Word).
        :param score_border: граница определния рифмы, чем выше, тем строже совпадение.
        :param syllable_number_border: ограничение на номер слога с конца, на который падает ударение.
        :param word_form_vocabulary: словарь словоформ.
        :return result: является рифмой или нет.
        """
        if word_form_vocabulary is not None:
            lemma1 = word_form_vocabulary.get_word_form_by_text(word1.text.lower()).lemma.lower()
            lemma2 = word_form_vocabulary.get_word_form_by_text(word2.text.lower()).lemma.lower()
            if lemma1 == lemma2:
                return False
        features1 = Rhymes.__get_rhyme_profile(word1)
        features2 = Rhymes.__get_rhyme_profile(word2)
        count_equality = 0
        for i, ch1 in enumerate(features1[1]):
            for j in range(i, len(features2[1])):
                ch2 = features2[1][j]
                if ch1 != ch2:
                    continue
                count_equality += 1
                if ch1 in VOWELS:
                    count_equality += 2
        if features1[2] == features2[2] and features1[2] != '' and features2[2] != '':
            count_equality += 2
        elif features1[3] == features2[3] and features1[3] != '' and features2[3] != '':
            count_equality += 1
        return features1[0] == features2[0] and count_equality >= score_border and \
               features1[0] <= syllable_number_border

    @staticmethod
    def __get_rhyme_profile(word: StressedWord) -> Tuple[int, str, str, str]:
        """
        Получение профиля рифмовки (набора признаков для сопоставления).

        :param word: уже акцентуированное слово (Word).
        :return profile: профиль рифмовки.
        """
        # TODO: Переход на фонетическое слово, больше признаков.
        syllable_number = 0
        stressed_syllable = ''
        next_syllable = ''
        next_char = ''
        syllables = list(reversed(word.syllables))
        for i, syllable in enumerate(syllables):
            if syllable.stress == -1:
                continue
            if i != 0:
                next_syllable = syllables[i - 1].text
            stressed_syllable = syllables[i].text
            if syllable.stress + 1 < len(word.text):
                next_char = word.text[syllable.stress + 1]
            syllable_number = i
            break
        return syllable_number, stressed_syllable, next_syllable, next_char



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль трансформеров языковой модели.

import numpy as np

from rupo.generate.word_form_vocabulary import WordFormVocabulary
from gensim.models import KeyedVectors


class Transformer(object):
    """
    Трансоформер языковой модели.
    """
    def get_coefficients(self) -> np.array:
        raise NotImplementedError()

    def transform_model(self, model: np.array) -> np.array:
        return np.multiply(model, self.get_coefficients())


class DistributionalSemanticTransformer(Transformer):
    """
    Трансоформер на основе семантических векторов.
    """
    def __init__(self, vocabulary: WordFormVocabulary, key_word: str, vectors_filename: str):
        w2v = KeyedVectors.load_word2vec_format(vectors_filename, binary=True)
        coefficients = []
        for i in range(len(vocabulary.word_forms)):
            current_word = vocabulary.get_word_form_by_index(i).lemma
            if key_word in w2v.vocab and current_word in w2v.vocab:
                coefficients.append(1.0 + w2v.similarity(key_word, current_word)*10)
            else:
                coefficients.append(1.0)
        self.coefficients = np.array(coefficients)
        pass

    def get_coefficients(self):
        return self.coefficients




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для токенизатора.

import unittest

from rupo.main.tokenizer import Tokenizer, Token


class TestTokenizer(unittest.TestCase):
    def test_tokenizer(self):
        text1 = "О, когда-нибудь, когда?"
        self.assertEqual(Tokenizer.tokenize(text1), [
            Token('О', Token.TokenType.WORD, 0, 1),
            Token(',', Token.TokenType.PUNCTUATION, 1, 2),
            Token(' ', Token.TokenType.SPACE, 2, 3),
            Token('когда-нибудь', Token.TokenType.WORD, 3, 15),
            Token(',', Token.TokenType.PUNCTUATION, 15, 16),
            Token(' ', Token.TokenType.SPACE, 16, 17),
            Token('когда', Token.TokenType.WORD, 17, 22),
            Token('?', Token.TokenType.PUNCTUATION, 22, 23)])



class Phonemes:
    VOWELS = ["i", "y", "ɪ", "ʏ", "ɨ", "ʉ", "ʊ", "ɯ", "u", "e", "ø", "ɘ", "ɵ", "ɤ", "o",
              "ə", "ɛ", "œ", "ɜ", "ɞ", "ʌ", "ɔ", "æ", "ɐ", "a", "ɶ", "ä", "ɑ", "ɒ", "ɝ",
              "ɚ"]
    APPROXIMANT_CONSONANTS = ["ʋ", "ɹ", "ɻ", "j", "ɰ", "ʍ", "w", "ɥ", "ʕ", "ʁ",
                              "l", "ɭ", "ʟ", "ʎ", "ɫ"]
    NASAL_CONSONANTS = ["m", "ɱ", "n", "ɳ", "ɲ", "ŋ", "ɴ", ]
    FLAP_OR_TAP_CONSONANTS = ["ⱱ", "ɾ", "ɽ", "ɺ"]
    IMPLOSIVE_CONSONANTS = ["ɓ", "ɗ", "ʄ", "ɠ", "ʛ"]
    CLICK_CONSONANTS = ["ʘ", "ǀ", "ǃ", "ǂ", "ǁ"]
    AFFRICATE_CONSONANTS = ["ʦ", "ʣ", "ʧ", "ʤ", "ʨ", "ʥ"]
    STOP_CONSONANTS = ["p", "b", "t", "d", "ʈ", "ɖ", "c", "ɟ", "k", "ɡ", "q",
                       "ɢ", "ʡ", "ʔ"]
    FRICATIVE_CONSONANTS = ["s", "z", "ʃ", "ʒ", "ʂ", "ʐ", "ɕ", "ʑ", "ɸ", "β",
                            "f", "v", "θ", "ð", "ç", "ʝ", "x", "ɣ", "χ", "ħ",
                            "h", "ɦ", "ɬ", "ɮ", "ʩ"]
    TRILL_CONSONANTS = ["ʙ", "r", "ʀ", "ʜ", "ʢ"]
    DIACRITICS = ["ʲ", "̯"]
    SUPRASEGMENTALS = ["ː"]
    LIGATURES = {
        "t͡s": "ʦ",
        "d͡z": "ʣ",
        "t͡ʃ": "ʧ",
        "d͡ʒ": "ʤ",
        "t͡ɕ": "ʨ",
        "d͡ʑ": "ʥ"
    }

    @staticmethod
    def get_all():
        l = [" "] + Phonemes.VOWELS + Phonemes.NASAL_CONSONANTS + Phonemes.STOP_CONSONANTS + \
            Phonemes.FRICATIVE_CONSONANTS + Phonemes.AFFRICATE_CONSONANTS + Phonemes.APPROXIMANT_CONSONANTS + \
            Phonemes.FLAP_OR_TAP_CONSONANTS + Phonemes.TRILL_CONSONANTS + Phonemes.CLICK_CONSONANTS + \
            Phonemes.IMPLOSIVE_CONSONANTS + Phonemes.DIACRITICS + Phonemes.SUPRASEGMENTALS
        return l

    @staticmethod
    def clean(phonemes: str) -> str:
        for letters, ch in Phonemes.LIGATURES.items():
            phonemes = phonemes.replace(letters, ch)
        clean = ""
        alphabet = Phonemes.get_all()
        for i, ch in enumerate(phonemes):
            if ch == "ː":
                if phonemes[i-1] in Phonemes.VOWELS:
                    clean += phonemes[i-1]
                continue
            if ch in alphabet + ["'", "ˌ"]:
                clean += ch
        j_positions = [i for i in range(len(clean)) if clean.startswith("ɪ̯", i)]
        offset = 0
        for pos in j_positions:
            if not (pos-1 >= 0 and clean[pos-1] not in Phonemes.VOWELS
                    and pos+2 < len(clean) and clean[pos+2] not in Phonemes.VOWELS):
                clean = clean[:pos-offset] + "j" + clean[pos-offset+2:]
                offset += 1
        return clean

    @staticmethod
    def get_sonority(phonemes: str):
        inverse_sonority_levels = [(Phonemes.VOWELS, 0), (Phonemes.APPROXIMANT_CONSONANTS, 1),
                                   (Phonemes.NASAL_CONSONANTS, 2), (Phonemes.FLAP_OR_TAP_CONSONANTS, 2),
                                   (Phonemes.IMPLOSIVE_CONSONANTS, 2), (Phonemes.STOP_CONSONANTS, 3),
                                   (Phonemes.CLICK_CONSONANTS, 3), (Phonemes.AFFRICATE_CONSONANTS, 4),
                                   (Phonemes.FRICATIVE_CONSONANTS, 5), (Phonemes.TRILL_CONSONANTS, 2)]
        sonority = []
        for ch in phonemes:
            for chars, level in inverse_sonority_levels:
                if ch in chars:
                    sonority.append(level)
            if ch in Phonemes.DIACRITICS or ch in Phonemes.SUPRASEGMENTALS:
                sonority.append(-1)
        return sonority

    @staticmethod
    def get_syllables(phonemes: str):
        sonority = Phonemes.get_sonority(phonemes)
        vowels_positions = [i for i, level in enumerate(sonority) if level == 0]
        intervals = [(vowels_positions[i]+1, vowels_positions[i+1]) for i in range(len(vowels_positions)-1)]
        borders = []
        for begin, end in intervals:
            sonority_interval = sonority[begin: end]
            border = begin
            if len(sonority_interval) != 0:
                border += len(sonority_interval) - sonority_interval[-1::-1].index(max(sonority_interval)) - 1
            borders.append(border)
        if len(borders) == 0:
            borders = [0]
        if borders[0] != 0:
            borders = [0] + borders
        if borders[-1] != len(phonemes):
            borders += [len(phonemes)]
        return [phonemes[borders[i]:borders[i+1]] for i in range(len(borders)-1)]


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Классификатор метра.

from collections import OrderedDict
from typing import List, Dict, Tuple
import jsonpickle
import logging

from rupo.main.markup import Line, Markup
from rupo.util.mixins import CommonMixin
from rupo.metre.pattern_analyzer import PatternAnalyzer
from rupo.util.preprocess import get_first_vowel_position
from rupo.util.timeit import timeit


class StressCorrection(CommonMixin):
    """
    Исправление ударения.
    """
    def __init__(self, line_number: int, word_number: int, syllable_number: int,
                 word_text: str, stress: int) -> None:
        """
        :param line_number: номер строки.
        :param word_number: номер слова.
        :param syllable_number: номер слога.
        :param word_text: текст слова.
        :param stress: позиция ударения (с 0).
        """
        self.line_number = line_number
        self.word_number = word_number
        self.syllable_number = syllable_number
        self.word_text = word_text
        self.stress = stress


class ClassificationResult(CommonMixin):
    """
    Результат классификации стихотворения по метру.
    """
    def __init__(self, count_lines: int=0) -> None:
        """
        :param count_lines: количество строк.
        """
        self.metre = None
        self.count_lines = count_lines
        self.errors_count = {k: 0 for k in MetreClassifier.metres.keys()}  # type: Dict[str, int]
        self.corrections = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[StressCorrection]]
        self.resolutions = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[StressCorrection]]
        self.additions = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[StressCorrection]]

    def get_metre_errors_count(self):
        """
        :return: получить количество ошибок на заданном метре.
        """
        return self.errors_count[self.metre]

    def to_json(self):
        """
        :return: сериализация в json.
        """
        return jsonpickle.encode(self)

    @staticmethod
    def str_corrections(collection: List[StressCorrection]) -> str:
        """
        :param collection: список исправлений.
        :return: его строковое представление.
        """
        return"\n".join([str((item.word_text, item.syllable_number)) for item in collection])

    def __str__(self):
        st = "Метр: " + str(self.metre) + "\n"
        st += "Снятая омография: \n" + ClassificationResult.str_corrections(self.resolutions[self.metre]) + "\n"
        st += "Неправильные ударения: \n" + ClassificationResult.str_corrections(self.corrections[self.metre]) + "\n"
        st += "Новые ударения: \n" + ClassificationResult.str_corrections(self.additions[self.metre]) + "\n"
        return st


class ErrorsTableRecord:
    def __init__(self, strong_errors, weak_errors, pattern, failed=False):
        self.strong_errors = strong_errors
        self.weak_errors = weak_errors
        self.pattern = pattern
        self.failed = failed

    def __str__(self):
        return self.pattern + " " + str(self.strong_errors) + " " + str(self.weak_errors)

    def __repr__(self):
        return self.__str__()


class ErrorsTable:
    def __init__(self, num_lines):
        self.data = {}
        self.num_lines = num_lines
        self.coef = OrderedDict(
            [("iambos", 0.3),
             ("choreios", 0.3),
             ("daktylos", 0.4),
             ("amphibrachys", 0.4),
             ("anapaistos", 0.4),
             ("dolnik3", 0.5),
             ("dolnik2", 0.5),
             ("taktovik3", 6.0),
             ("taktovik2", 6.0)
             ])
        self.sum_coef = OrderedDict(
            [("iambos", 0.0),
             ("choreios", 0.0),
             ("daktylos", 0.0),
             ("amphibrachys", 0.0),
             ("anapaistos", 0.0),
             ("dolnik3", 0.035),
             ("dolnik2", 0.035),
             ("taktovik3", 0.10),
             ("taktovik2", 0.10)
             ])
        for metre_name in MetreClassifier.metres.keys():
            self.data[metre_name] = [ErrorsTableRecord(0, 0, "") for _ in range(num_lines)]

    def add_record(self, metre_name, line_num, strong_errors, weak_errors, pattern, failed=False):
        self.data[metre_name][line_num] = ErrorsTableRecord(strong_errors, weak_errors, pattern, failed)

    def get_best_metre(self):
        for l in range(self.num_lines):
            strong_sum = 0
            weak_sum = 0
            for metre_name in self.data.keys():
                strong_sum += self.data[metre_name][l].strong_errors
                weak_sum += self.data[metre_name][l].weak_errors
            for metre_name, column in self.data.items():
                if strong_sum != 0:
                    column[l].strong_errors = column[l].strong_errors / float(strong_sum)
                if weak_sum != 0:
                    column[l].weak_errors = column[l].weak_errors / float(weak_sum)
        sums = dict()
        for metre_name in self.data.keys():
            sums[metre_name] = (0, 0)
        for metre_name, column in self.data.items():
            strong_sum = 0
            weak_sum = 0
            for l in range(self.num_lines):
                strong_sum += column[l].strong_errors
                weak_sum += column[l].weak_errors
            sums[metre_name] = (strong_sum, weak_sum)
        for metre_name, pair in sums.items():
            sums[metre_name] = self.sum_coef[metre_name] + (pair[0] + pair[1] / 2.0) * self.coef[metre_name] / self.num_lines
        logging.debug(sums)
        return min(sums, key=sums.get)


class MetreClassifier(object):
    """
    Классификатор, считает отклонения от стандартных шаблонов ритма(метров).
    """
    metres = OrderedDict(
        [("iambos", '(us)*(uS)(U)?(U)?'),
         ("choreios", '(su)*(S)(U)?(U)?'),
         ("daktylos", '(suu)*(S)(U)?(U)?'),
         ("amphibrachys", '(usu)*(uS)(U)?(U)?'),
         ("anapaistos",  '(uus)*(uuS)(U)?(U)?'),
         ("dolnik3", '(u)?(u)?((su)(u)?)*(S)(U)?(U)?'),
         ("dolnik2", '(u)?(u)?((s)(u)?)*(S)(U)?(U)?'),
         ("taktovik3", '(u)?(u)?((su)(u)?(u)?)*(S)(U)?(U)?'),
         ("taktovik2", '(u)?(u)?((s)(u)?(u)?)*(S)(U)?(U)?')
         ])

    border_syllables_count = 20

    @staticmethod
    @timeit
    def classify_metre(markup):
        """
        Классифицируем стихотворный метр.

        :param markup: разметка.
        :return: результат классификации.
        """
        result = ClassificationResult(len(markup.lines))
        num_lines = len(markup.lines)
        errors_table = ErrorsTable(num_lines)
        for l, line in enumerate(markup.lines):
            for metre_name, metre_pattern in MetreClassifier.metres.items():
                line_syllables_count = sum([len(word.syllables) for word in line.words])

                # Строчки длиной больше border_syllables_count слогов не обрабатываем.
                if line_syllables_count > MetreClassifier.border_syllables_count or line_syllables_count == 0:
                    continue
                error_border = 7
                if metre_name == "dolnik2" or metre_name == "dolnik3":
                    error_border = 3
                if metre_name == "taktovik2" or metre_name == "taktovik3":
                    error_border = 2
                pattern, strong_errors, weak_errors, analysis_errored = \
                    PatternAnalyzer.count_errors(MetreClassifier.metres[metre_name],
                                                 MetreClassifier.__get_line_pattern(line),
                                                 error_border)
                if analysis_errored or len(pattern) == 0:
                    errors_table.add_record(metre_name, l, strong_errors, weak_errors, pattern, True)
                    continue
                corrections = MetreClassifier.__get_line_pattern_matching_corrections(line, l, pattern)[0]
                accentuation_errors = len(corrections)
                strong_errors += accentuation_errors
                errors_table.add_record(metre_name, l, strong_errors, weak_errors, pattern)
        result.metre = errors_table.get_best_metre()

        # Запомним все исправления.
        for l, line in enumerate(markup.lines):
            pattern = errors_table.data[result.metre][l].pattern
            failed = errors_table.data[result.metre][l].failed
            if failed or len(pattern) == 0:
                continue
            corrections, resolutions, additions =\
                MetreClassifier.__get_line_pattern_matching_corrections(line, l, pattern)
            result.corrections[result.metre] += corrections
            result.resolutions[result.metre] += resolutions
            result.additions[result.metre] += additions
            result.errors_count[result.metre] += len(corrections)
        return result

    @staticmethod
    def __get_line_pattern(line: Line) -> str:
        """
        Сопоставляем строку шаблону, считаем ошибки.

        :param line: строка.
        :return: количество ошибок
        """
        pattern = ""
        for w, word in enumerate(line.words):
            if len(word.syllables) == 0:
                pattern += "U"
            else:
                for syllable in word.syllables:
                    if syllable.stress != -1:
                        pattern += "S"
                    else:
                        pattern += "U"
        return pattern

    @staticmethod
    def __get_line_pattern_matching_corrections(line: Line, line_number: int, pattern: str) \
            -> Tuple[List[StressCorrection], List[StressCorrection], List[StressCorrection]]:
        """
        Ударения могут приходиться на слабое место,
        если безударный слог того же слова не попадает на икт. Иначе - ошибка.

        :param line: строка.
        :param line_number: номер строки.
        :param pattern: шаблон.
        :return: ошибки, дополнения и снятия
        """
        corrections = []
        resolutions = []
        additions = []
        number_in_pattern = 0
        for w, word in enumerate(line.words):
            # Игнорируем слова длиной меньше 2 слогов.
            if len(word.syllables) == 0:
                continue
            if len(word.syllables) == 1:
                if pattern[number_in_pattern].lower() == "s" and word.syllables[0].stress == -1:
                    additions.append(StressCorrection(line_number, w, 0, word.text, word.syllables[0].vowel()))
                number_in_pattern += len(word.syllables)
                continue
            stress_count = word.count_stresses()
            for syllable in word.syllables:
                if stress_count == 0 and pattern[number_in_pattern].lower() == "s":
                    # Ударений нет, ставим такое, какое подходит по метру. Возможно несколько.
                    additions.append(StressCorrection(line_number, w, syllable.number, word.text, syllable.vowel()))
                elif pattern[number_in_pattern].lower() == "u" and syllable.stress != -1:
                    # Ударение есть и оно падает на этот слог, при этом в шаблоне безударная позиция.
                    # Найдём такой слог, у которого в шаблоне ударная позиция. Это и есть наше исправление.
                    for other_syllable in word.syllables:
                        other_number_in_pattern = other_syllable.number - syllable.number + number_in_pattern
                        if syllable.number == other_syllable.number or pattern[other_number_in_pattern].lower() != "s":
                            continue
                        ac = StressCorrection(line_number, w, other_syllable.number, word.text, other_syllable.vowel())
                        if stress_count == 1 and other_syllable.stress == -1:
                            corrections.append(ac)
                        else:
                            resolutions.append(ac)
                number_in_pattern += 1
        return corrections, resolutions, additions

    @staticmethod
    def get_improved_markup(markup: Markup, result: ClassificationResult) -> Markup:
        """
        Улучшаем разметку после классификации метра.

        :param markup: начальная разметка.
        :param result: результат классификации.
        :return: улучшенная разметка.
        """
        for pos in result.corrections[result.metre] + result.resolutions[result.metre]:
            syllables = markup.lines[pos.line_number].words[pos.word_number].syllables
            for i, syllable in enumerate(syllables):
                syllable.stress = -1
                if syllable.number == pos.syllable_number:
                    syllable.stress = syllable.begin + get_first_vowel_position(syllable.text)
        for pos in result.additions[result.metre]:
            syllable = markup.lines[pos.line_number].words[pos.word_number].syllables[pos.syllable_number]
            syllable.stress = syllable.begin + get_first_vowel_position(syllable.text)

        return markup

    @staticmethod
    def improve_markup(markup: Markup) -> \
            Tuple[Markup, ClassificationResult]:
        """
        Улучшение разметки метрическим классификатором.

        :param markup: начальная разметка.
        """
        result = MetreClassifier.classify_metre(markup)
        improved_markup = MetreClassifier.get_improved_markup(markup, result)
        return improved_markup, result



# -*- coding: utf-8 -*-
# Автор: Даниил Анастасьев
# Описание: Загрузка словарей из корпуса.

from collections import Counter
from typing import List, Tuple
import pickle
import os

import pymorphy2
from russian_tagsets import converters

from rupo.generate.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.tqdm_open import tqdm_open


class WordVocabulary:
    def __init__(self, dump_filename):
        self.dump_filename = dump_filename
        self.words = []
        self.word_to_index = {}
        self.counter = Counter()
        if os.path.exists(self.dump_filename):
            self.load()

    def add_word(self, word):
        if word in self.word_to_index:
            self.counter[word] += 1
        else:
            self.words.append(word)
            self.counter[word] = 1
            self.word_to_index[word] = len(self.words) - 1

    def sort(self):
        self.words = []
        self.word_to_index = {}
        for word, _ in self.counter.most_common():
            self.words.append(word)
            self.word_to_index[word] = len(self.words) - 1

    def is_empty(self):
        return len(self.words) == 0

    def save(self) -> None:
        """
        Сохранение словаря.
        """
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def load(self) -> None:
        """
        Загрузка словаря.
        """
        with open(self.dump_filename, "rb") as f:
            vocab = pickle.load(f)
            self.__dict__.update(vocab.__dict__)


class Loader(object):
    """
    Класс для построения GrammemeVectorizer и WordFormVocabulary по корпусу
    """
    def __init__(self, gram_dump_path, word_dump_path):
        self.grammeme_vectorizer = GrammemeVectorizer(gram_dump_path)
        self.word_vocabulary = WordVocabulary(word_dump_path)
        self.morph = pymorphy2.MorphAnalyzer()

    def parse_corpora(self, filenames: List[str]) -> Tuple[GrammemeVectorizer, WordVocabulary]:
        """
        Построить WordFormVocabulary, GrammemeVectorizer по корпусу

        :param filenames: пути к файлам корпуса.
        """
        for filename in filenames:
            with tqdm_open(filename, encoding="utf-8") as f:
                for line in f:
                    if line == "\n":
                        continue
                    self.__process_line(line)

        self.grammeme_vectorizer.init_possible_vectors()
        return self.grammeme_vectorizer, self.word_vocabulary

    def __process_line(self, line: str) -> None:
        text, lemma, pos_tag, grammemes = line.strip().split("\t")[:4]
        self.word_vocabulary.add_word(text)
        self.grammeme_vectorizer.add_grammemes(pos_tag, grammemes)
        to_ud = converters.converter('opencorpora-int', 'ud14')
        for parse in self.morph.parse(text):
            ud_tag = to_ud(str(parse.tag), text)
            pos = ud_tag.split()[0]
            gram = ud_tag.split()[1].split("|")
            dropped = ["Animacy", "Aspect", "NumType"]
            gram = [grammem for grammem in gram if sum([drop in grammem for drop in dropped]) == 0]
            gram = "|".join(gram)
            self.grammeme_vectorizer.add_grammemes(pos, gram)


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для выравнивания g2pц.

import unittest

from rupo.g2p.aligner import Aligner
from rupo.settings import RU_G2P_DICT_PATH


class TestAligner(unittest.TestCase):
    def test_aligner(self):
        aligner = Aligner()
        with open(RU_G2P_DICT_PATH, 'r', encoding='utf-8') as r:
            lines = r.readlines()[:50]
            pairs = [tuple(line.strip().split("\t")) for line in lines]
            for g, p in pairs:
                print(aligner.align(g, p))




# import numpy as np
# from typing import List, Tuple
# from collections import Counter
# from itertools import islice
# from keras.models import model_from_json,  Model, load_model
# from keras.layers import Input, Embedding, Dense, LSTM, SpatialDropout1D, Masking, \
#     BatchNormalization, Activation, concatenate, Bidirectional, TimeDistributed, Dropout
# from russian_tagsets import converters
#
# from rupo.generate.word_form import WordForm
# from rupo.generate.word_form_vocabulary import WordFormVocabulary
# from rupo.generate.grammeme_vectorizer import GrammemeVectorizer
# from rupo.generate.tqdm_open import tqdm_open
# from rupo.settings import GENERATOR_LSTM_MODEL_PATH, GENERATOR_WORD_FORM_VOCAB_PATH, GENERATOR_GRAM_VECTORS
# import pymorphy2
# from rupo.morph.loader import WordVocabulary, Loader
#
#
# class BatchGenerator:
#     """
#     Генератор наборов примеров для обучения.
#     """
#
#     def __init__(self, filenames: List[str], batch_size: int,
#                  input_size: int, context_maxlen: int, word_vocabulary: WordVocabulary,
#                  grammeme_vectorizer: GrammemeVectorizer):
#         """
#         :param filenames: имена файлов с морфоразметкой.
#         :param batch_size: размер набора семплов.
#         :param word_form_vocabulary: словарь словофрм.
#         :param grammeme_vectorizer: векторизатор граммем.
#         """
#         self.filenames = filenames  # type: List[str]
#         self.batch_size = batch_size  # type: int
#         self.input_size = input_size  # type: int
#         self.context_maxlen = context_maxlen  # type: int
#         self.word_vocabulary = word_vocabulary  # type: WordVocabulary
#         self.grammeme_vectorizer = grammeme_vectorizer  # type: GrammemeVectorizer
#         self.morph = pymorphy2.MorphAnalyzer()
#
#     def __to_tensor(self, sentences: List[List[WordForm]]) -> Tuple[np.array, np.array, np.array]:
#         n_samples = len(sentences)
#         max_len = self.context_maxlen
#         words = np.zeros((n_samples, max_len), dtype=np.int)
#         grammemes = np.zeros((n_samples, max_len, self.grammeme_vectorizer.size()), dtype=np.int)
#         y = np.zeros((n_samples, self.grammeme_vectorizer.grammemes_count()), dtype=np.int)
#         for i in range(n_samples):
#             sentence = sentences[i]
#             texts = [x.text for x in sentence]
#             words[i, -len(sentence):] = \
#                 [min(self.word_vocabulary.word_to_index[word], self.input_size) for word in texts]
#
#             to_ud = converters.converter('opencorpora-int', 'ud14')
#             gram_vectors = []
#             for text in texts:
#                 gram_value_indices = np.zeros(self.grammeme_vectorizer.size())
#                 for parse in self.morph.parse(text):
#                     ud_tag = to_ud(str(parse.tag), text)
#                     pos = ud_tag.split()[0]
#                     gram = ud_tag.split()[1].split("|")
#                     gram_value_indices[self.grammeme_vectorizer.get_index_by_name(pos + "#" + gram)] = 1
#                 gram_vectors.append(gram_value_indices)
#             grammemes[i, -len(sentence):] = gram_vectors
#
#             y[i, -len(sentence):] = [x.gram_vector_index for x in sentence]
#         return words, grammemes, y
#
#     def __iter__(self):
#         """
#         Получение очередного батча.
#
#         :return: индексы словоформ, грамматические векторы, ответы-индексы.
#         """
#         sentences = [[]]
#         for filename in self.filenames:
#             with tqdm_open(filename, encoding='utf-8') as f:
#                 for line in f:
#                     line = line.strip()
#                     if len(line) == 0:
#                         sentences.append([])
#                     else:
#                         word, lemma, pos, tags = line.split('\t')[:4]
#                         word, lemma = word.lower(), lemma.lower() + '_' + pos
#                         tags = "|".join(sorted(tags.split("|")))
#                         gram_vector_index = self.grammeme_vectorizer.name_to_index[pos + "#" + tags]
#                         sentences[-1].append(WordForm(lemma, gram_vector_index, word))
#                     if len(sentences) >= self.batch_size:
#                         yield self.__to_tensor(sentences)
#                         sentences = [[]]
#
#
# class LSTMMorphoAnalysis:
#     def __init__(self, input_size: int=25000, external_batch_size: int=10000, nn_batch_size: int=768,
#                  context_maxlen: int=10, lstm_units=368, embeddings_dimension: int=250,
#                  grammeme_dense_units: Tuple[int]=(100, 50), dense_units: int=256):
#         self.input_size = input_size # type: int
#         self.external_batch_size = external_batch_size  # type: int
#         self.nn_batch_size = nn_batch_size  # type: int
#         self.context_maxlen = context_maxlen  # type: int
#         self.word_form_vocabulary = None  # type: WordFormVocabulary
#         self.grammeme_vectorizer = None  # type: GrammemeVectorizer
#         self.lstm_units = lstm_units  # type: int
#         self.embeddings_dimension = embeddings_dimension  # type: int
#         self.grammeme_dense_units = grammeme_dense_units  # type: List[int]
#         self.dense_units = dense_units  # type: int
#         self.model = None  # type: Model
#
#     def prepare(self, filenames: List[str], word_vocab_dump_path, gram_dump_path) -> None:
#         """
#         Подготовка векторизатора грамматических значений и словаря словоформ по корпусу.
#
#         :param filenames: имена файлов с морфоразметкой.
#         :param word_form_vocab_dump_path: путь к дампу словаря словоформ.
#         :param gram_dump_path: путь к векторам грамматических значений.
#         """
#         self.grammeme_vectorizer = GrammemeVectorizer(gram_dump_path)
#         self.word_vocabulary = WordVocabulary(word_vocab_dump_path)
#         if self.grammeme_vectorizer.is_empty() or self.word_vocabulary.is_empty():
#             loader = Loader(gram_dump_path, word_vocab_dump_path)
#             self.grammeme_vectorizer, self.word_vocabulary = loader.parse_corpora(filenames)
#             self.grammeme_vectorizer.save()
#             self.word_vocabulary.save()
#
#     # def load(self, model_filename: str) -> None:
#     #     """
#     #     Загрузка модели.
#     #
#     #     :param model_filename: файл с моделью.
#     #     """
#     #     self.model = load_model(model_filename)
#     #
#     # def load_with_weights(self, json_filename: str, weights_filename: str) -> None:
#     #     """
#     #     Загрузка модели из json описания и файла с весами.
#     #
#     #     :param json_filename: json описание.
#     #     :param weights_filename: файл с весам.
#     #     """
#     #     json_string = open(json_filename, 'r', encoding='utf8').readline()
#     #     self.model = model_from_json(json_string)
#     #     self.model.load_weights(weights_filename)
#     #
#     # def build(self):
#     #     """
#     #     Описание модели.
#     #     """
#     #     # Вход лемм
#     #     lemmas = Input(shape=(None,), name='lemmas')
#     #     lemmas_embedding = Embedding(self.input_size + 1, self.embeddings_dimension, name='embeddings')(lemmas)
#     #     lemmas_embedding = SpatialDropout1D(.3)(lemmas_embedding)
#     #
#     #     # Вход граммем
#     #     grammemes_input = Input(shape=(None, self.grammeme_vectorizer.grammemes_count()), name='grammemes')
#     #     grammemes_layer = Masking(mask_value=0.)(grammemes_input)
#     #     for grammeme_dense_layer_units in self.grammeme_dense_units:
#     #         grammemes_layer = Dense(grammeme_dense_layer_units, activation='relu')(grammemes_layer)
#     #
#     #     layer = concatenate([lemmas_embedding, grammemes_layer], name="LSTM_input")
#     #     layer = Bidirectional(LSTM(self.lstm_units, dropout=.2, recurrent_dropout=.2,
#     #                                return_sequences=True, name='LSTM_1'))(layer)
#     #     layer = Bidirectional(LSTM(self.lstm_units, dropout=.2, recurrent_dropout=.2,
#     #                                return_sequences=False, name='LSTM_2'))(layer)
#     #     layer = TimeDistributed(Dropout(.2))(layer)
#     #     layer = TimeDistributed(Dense(self.dense_units))(layer)
#     #
#     #     layer = Dense(self.dense_units)(layer)
#     #     layer = BatchNormalization()(layer)
#     #     layer = Activation('relu')(layer)
#     #
#     #     output = Dense(self.softmax_size + 1, activation='softmax')(layer)
#     #
#     #     self.model = Model(inputs=[lemmas, grammemes_input], outputs=[output])
#     #     self.model.compile(loss='sparse_categorical_crossentropy', optimizer='adam')
#     #     print(self.model.summary())
#     #
#     # @staticmethod
#     # def __get_validation_data(batch_generator, size):
#     #     """
#     #     Берет первые size батчей из batch_generator для валидационной выборки
#     #     """
#     #     lemmas_list, grammemes_list, y_list = [], [], []
#     #     for lemmas, grammemes, y in islice(batch_generator, size):
#     #         lemmas_list.append(lemmas)
#     #         grammemes_list.append(grammemes)
#     #         y_list.append(y)
#     #     return np.vstack(lemmas_list), np.vstack(grammemes_list), np.hstack(y_list)
#     #
#     # def train(self, filenames: List[str], validation_size: int = 5,
#     #           validation_verbosity: int = 5, dump_model_freq: int = 10) -> None:
#     #     """
#     #     Обучение модели.
#     #
#     #     :param filenames: имена файлов с морфоразметкой.
#     #     :param validation_size: размер val выборки.
#     #     :param validation_verbosity: каждый validation_verbosity-шаг делается валидация.
#     #     :param dump_model_freq: каждый dump_model_freq-шаг сохраняется модель.
#     #     """
#     #     batch_generator = BatchGenerator(filenames,
#     #                                      batch_size=self.external_batch_size,
#     #                                      embedding_size=self.embedding_size,
#     #                                      softmax_size=self.softmax_size,
#     #                                      sentence_maxlen=self.sentence_maxlen,
#     #                                      word_form_vocabulary=self.word_form_vocabulary,
#     #                                      grammeme_vectorizer=self.grammeme_vectorizer)
#     #
#     #     lemmas_val, grammemes_val, y_val = LSTMGenerator.__get_validation_data(batch_generator, validation_size)
#     #     for big_epoch in range(0, 1000):
#     #         print('------------Big Epoch {}------------'.format(big_epoch))
#     #         for epoch, (lemmas, grammemes, y) in enumerate(batch_generator):
#     #             # if epoch < 2480:
#     #             #     continue
#     #             if epoch < validation_size:
#     #                 continue
#     #             self.model.fit([lemmas, grammemes], y, batch_size=self.nn_batch_size, epochs=1, verbose=2)
#     #
#     #             if epoch != 0 and epoch % validation_verbosity == 0:
#     #                 print('val loss:',
#     #                       self.model.evaluate([lemmas_val, grammemes_val],
#     #                                           y_val, batch_size=self.nn_batch_size * 2, verbose=0))
#     #
#     #             indices = [self.word_form_vocabulary.get_sequence_end_index()]
#     #             for _ in range(10):
#     #                 indices.append(self._sample(self.predict(indices)))
#     #             sentence = [self.word_form_vocabulary.get_word_form_by_index(index) for index in indices]
#     #             print('Sentence', str(big_epoch), str(epoch), end=': ')
#     #             for word in sentence[::-1]:
#     #                 print(word.text, end=' ')
#     #             print()
#     #
#     #             if epoch != 0 and epoch % dump_model_freq == 0:
#     #                 self.model.save(GENERATOR_LSTM_MODEL_PATH)
#     #
#     # def predict(self, word_indices: List[int]) -> np.array:
#     #     """
#     #     Предсказание вероятностей следующего слова.
#     #
#     #     :param word_indices: индексы предыдущих слов.
#     #     :return: проекция языковой модели (вероятности следующего слова).
#     #     """
#     #     if len(word_indices) == 0:
#     #         return np.full(self.softmax_size, 1.0 / self.softmax_size, dtype=np.float)
#     #     cur_sent = [self.word_form_vocabulary.get_word_form_by_index(ind) for ind in word_indices]
#     #     x_lemmas = np.zeros((1, len(cur_sent)))
#     #     x_grammemes = np.zeros((1, len(cur_sent), self.grammeme_vectorizer.grammemes_count()))
#     #     for index, word in enumerate(cur_sent):
#     #         x_lemmaslf.word_form_vocabulary.get_lemma_index(word)
#     #         x_grammemes[0, index][0, index] = se = self.grammeme_vectorizer.vectors[word.gram_vector_index]
#     #     prob = self.model.predict([x_lemmas, x_grammemes], verbose=0)[0]
#     #     return prob
#     #
#     # @staticmethod
#     # def _sample(prob: np.array, temperature: float = 1.0) -> int:
#     #     """
#     #     Выбор слова по набору вероятностей с заданной температурой (распределение Больцмана).
#     #
#     #     :param prob: вероятности.
#     #     :param temperature: температура.
#     #     :return: индекс итогового слова.
#     #     """
#     #     prob = prob[:-1]  # Для исключения неизвестных слов.
#     #     prob = np.log(prob) / temperature
#     #     prob = np.exp(prob) / np.sum(np.exp(prob))
#     #     return np.random.choice(len(prob), p=prob)


# -*- coding: utf-8 -*-
# Автор: Даниил Анастасьев
# Описание: Загрузка словарей из корпуса.

from collections import defaultdict, Counter
from typing import List, Tuple, Dict, Set
import sys

from rupo.generate.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.word_form_vocabulary import WordFormVocabulary, SEQ_END, SEQ_END_WF
from rupo.generate.word_form import WordForm, LemmaCase
from rupo.generate.tqdm_open import tqdm_open


class CorporaInformationLoader(object):
    """
    Класс для построения GrammemeVectorizer и WordFormVocabulary по корпусу
    """
    def __init__(self):
        self.grammeme_vectorizer = GrammemeVectorizer()
        self.word_form_vocabulary = WordFormVocabulary()
        self.lemma_to_word_forms = defaultdict(set)  # type: Dict[str, Set[WordForm]]
        self.lemma_case = {}
        self.lemma_counter = Counter()  # type: Counter

    def parse_corpora(self, filenames: List[str]) -> Tuple[WordFormVocabulary, GrammemeVectorizer]:
        """
        Построить WordFormVocabulary, GrammemeVectorizer по корпусу

        :param filenames: пути к файлам корпуса.
        """
        for filename in filenames:
            with tqdm_open(filename, encoding="utf-8") as f:
                for line in f:
                    if line == "\n":
                        continue
                    self.__process_line(line)

        self.__add_seq_end()
        self.grammeme_vectorizer.init_possible_vectors()
        self.word_form_vocabulary.init_by_vocabulary(self.lemma_counter, self.lemma_to_word_forms, self.lemma_case)
        self.word_form_vocabulary.lemma_indices[SEQ_END_WF] = 1
        return self.word_form_vocabulary, self.grammeme_vectorizer

    def __add_seq_end(self):
        self.lemma_to_word_forms[SEQ_END].add(SEQ_END_WF)
        self.lemma_case[SEQ_END] = SEQ_END_WF.case
        self.lemma_counter[SEQ_END] = sys.maxsize

    def __process_line(self, line: str) -> None:
        try:
            text, lemma, pos_tag, grammemes = line.strip().split("\t")[:4]
            lemma = lemma.lower() + '_' + pos_tag
            gram_vector_index = self.grammeme_vectorizer.add_grammemes(pos_tag, grammemes)
            self.lemma_to_word_forms[lemma].add(WordForm(lemma, gram_vector_index, text.lower()))
            self.lemma_counter[lemma] += 1
            self.__update_lemma_case(lemma, text)
        except ValueError:
            pass

    def __update_lemma_case(self, lemma: str, text: str) -> None:
        if lemma not in self.lemma_case:
            self.lemma_case[lemma] = LemmaCase.UPPER_CASE if text.isupper() else \
                              LemmaCase.PROPER_CASE if text[0].isupper() else LemmaCase.NORMAL_CASE
        elif self.lemma_case[lemma] == LemmaCase.UPPER_CASE:
            if not text.isupper():
                self.lemma_case[lemma] = LemmaCase.PROPER_CASE if text[0].isupper() else LemmaCase.NORMAL_CASE
        elif self.lemma_case[lemma] == LemmaCase.PROPER_CASE:
            if not text[0].isupper():
                self.lemma_case[lemma] = LemmaCase.NORMAL_CASE


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль марковских цепей.

import os
import pickle
import numpy as np
from collections import Counter, defaultdict
from typing import List, Dict, Tuple

from rupo.files.reader import Reader, FileType
from rupo.main.markup import Markup
from rupo.main.vocabulary import StressVocabulary
from rupo.generate.model_container import ModelContainer


class MarkovModelContainer(ModelContainer):
    """
    Марковские цепи.
    """
    def __init__(self, dump_filename: str, vocabulary: StressVocabulary, markup_dump_path: str=None,
                 n_poems: int=None, n_grams: int=2):
        self.n_grams = n_grams
        self.transitions = defaultdict(Counter)  # type: Dict[Tuple, Counter]
        self.vocabulary = vocabulary
        self.dump_filename = dump_filename

        # Делаем дамп модели для ускорения загрузки.
        if os.path.isfile(dump_filename):
            self.load()
        else:
            i = 0
            markups = Reader.read_markups(markup_dump_path, FileType.XML, is_processed=True)
            for markup in markups:
                self.add_markup(markup)
                i += 1
                if n_poems is not None and n_poems == i:
                    break
                if i % 500 == 0:
                    print(i)
            self.save()

    def save(self):
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self.transitions, f, pickle.HIGHEST_PROTOCOL)
        self.vocabulary.save()

    def load(self):
        with open(self.dump_filename, "rb") as f:
            self.transitions = pickle.load(f)
        self.vocabulary.load()

    def generate_chain(self, words: List[int]) -> Dict[Tuple, Counter]:
        """
        Генерация переходов в марковских цепях с учётом частотности.

        :param words: вершины цепи.
        :return: обновленные переходы.
        """
        for i in range(len(words) - self.n_grams + 1):
            current_words = tuple([words[j] for j in range(i, i+self.n_grams-1)])
            next_word = words[i+1]
            self.transitions[current_words][next_word] += 1
        return self.transitions

    def add_markup(self, markup: Markup) -> None:
        """
        Дополнение цепей на основе разметки.

        :param markup: разметка.
        """
        words = []
        for line in markup.lines:
            for word in line.words:
                try:
                    words.append(self.vocabulary.get_word_index(word.to_stressed_word()))
                except IndexError:
                    print("Can't find word in vocabulary.")
                    pass

        # Генерируем переходы.
        self.generate_chain(list(reversed(words)))

    def get_model(self, word_indices: List[int]) -> np.array:
        """
        Получение языковой модели.

        :param word_indices: индексы предыдущих слов.
        :return: языковая модель (распределение вероятностей для следующего слова).
        """
        l = self.vocabulary.size()
        if len(word_indices) < self.n_grams - 1:
            return np.full(self.vocabulary.size(), 1/l, dtype=np.float)
        prev_words = tuple([word_indices[-i] for i in range(1, self.n_grams)])
        if len(self.transitions[prev_words]) == 0:
            return np.full(self.vocabulary.size(), 1/l, dtype=np.float)
        else:
            transition = self.transitions[prev_words]
            s = sum(transition.values())
            model = np.zeros(self.vocabulary.size(), dtype=np.float)
            for index, p in transition.items():
                model[index] = p/s
            return model



# -*- coding: utf-8 -*-
# Авторы: Гусев Илья
# Описание: Контейнер языковой модели.

import numpy as np
from typing import List


class ModelContainer(object):
    """
    Контейнер языковой модели.
    """
    def get_model(self, word_indices: List[int]) -> np.array:
        """
        Получение проекции языковой модели.
        
        :param word_indices: индексы предыдущих слов.
        :return: вероятности следующего слова.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Рекуррентная сеть для получения ударений по фонемам.

import numpy as np
import os
import logging

from typing import List, Tuple

from sklearn.model_selection import train_test_split
from keras.models import Model, load_model
from keras.preprocessing import sequence
from keras.callbacks import EarlyStopping, ModelCheckpoint, Callback
from keras.layers import LSTM, Bidirectional, Dropout, Dense, TimeDistributed, Input, Embedding

from rupo.settings import RU_GRAPHEME_SET, RU_GRAPHEME_STRESS_PATH, DATA_DIR


class RNNGraphemeStressModel:
    def __init__(self, dict_path: str=None, word_max_length: int=30, language: str="ru", rnn=LSTM,
                 units: int=64, dropout: float=0.2, batch_size=2048, emb_dimension=30):
        self.rnn = rnn
        self.dropout = dropout  # type: float
        self.units = units  # type: int
        self.language = language  # type: str
        self.dict_path = dict_path  # type: str
        self.word_max_length = word_max_length  # type: int
        self.batch_size = batch_size
        self.emb_dimension = emb_dimension
        self.model = None
        if language == "ru":
            self.grapheme_set = RU_GRAPHEME_SET

    def build(self) -> None:
        """
        Построение модели. 
        """
        inp = Input(shape=(None,))

        emb = Embedding(len(self.grapheme_set), self.emb_dimension)(inp)
        encoded = Bidirectional(self.rnn(self.units, return_sequences=True, recurrent_dropout=self.dropout))(emb)
        encoded = Dropout(self.dropout)(encoded)
        decoded = TimeDistributed(Dense(self.units, activation="relu"))(encoded)
        predictions = TimeDistributed(Dense(3, activation="softmax"))(decoded)

        model = Model(inputs=inp, outputs=predictions)
        model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        print(model.summary())
        self.model = model

    def train(self, dir_name: str, enable_checkpoints: bool = False) -> None:
        """
        Обучение сети.

        :param dir_name: папка, в которую сохраняеются все весрии модели.
        :param enable_checkpoints: использовать ли чекпоинты.
        """
        # Подготовка данных
        x, y = self.__load_dict()
        x, y = self.__prepare_data(x, y)
        # Деление на выборки.
        x_train, x_val, y_train, y_val = train_test_split(x, y, test_size=0.2, random_state=42)
        x_test, x_val, y_test, y_val = train_test_split(x_val, y_val, test_size=0.5, random_state=42)
        # Основные раунды обучения.
        callbacks = [EarlyStopping(monitor='val_acc', patience=3)]  # type: List[Callback]
        if enable_checkpoints:
            checkpoint_name = os.path.join(dir_name, "checkpoint.hdf5")
            callbacks.append(ModelCheckpoint(checkpoint_name, monitor='val_loss'))
        self.model.fit(x_train, y_train, verbose=1, epochs=200, validation_data=(x_val, y_val),
                       callbacks=callbacks, batch_size=self.batch_size)
        # Рассчёт точности на test выборке.
        accuracy = self.model.evaluate(x_test, y_test)[1]
        # Расчёт WER на test выборке.
        wer = self.__evaluate_wer(x_test, y_test)[0]
        # Один раунд обучения на всём датасете.
        self.model.fit(x, y, verbose=1, epochs=1, batch_size=self.batch_size)
        # Сохранение модели.
        filename = "stress_{language}_{rnn}{units}_dropout{dropout}_acc{acc}_wer{wer}.h5"
        filename = filename.format(language=self.language, rnn=self.rnn.__name__,
                                   units=self.units, dropout=self.dropout, acc=int(accuracy * 100),
                                   wer=int(wer * 100))
        self.model.save(os.path.join(dir_name, filename))

    def predict(self, words: List[str]) -> List[List[int]]:
        """
        Предсказание ударений.

        :param words: слова. 
        :return: ударения.
        """
        x, y = self.__prepare_data(words, None)
        y = self.model.predict(x, verbose=0, batch_size=self.batch_size)
        answers = []
        for word in y:
            answer = []
            for ch_prob in word:
                i = int(np.argmax(ch_prob))
                answer.append(i)
            answers.append(answer)
        return answers

    def load(self, filename: str) -> None:
        self.model = load_model(filename)

    def __load_dict(self) -> Tuple[List[str], np.array]:
        """
        Парсинг словаря.

        :return: графические слова и ударения.
        """
        x = []
        y = []
        skipped = 0
        with open(self.dict_path, "r", encoding='utf-8') as f:
            for line in f:
                graphemes, primary, secondary = line.split("\t")
                graphemes = graphemes.lower()
                primary = [int(i) for i in primary.split(",") if i != '']
                secondary = [int(i) for i in secondary.strip().split(",") if i != '']
                if len(graphemes) > self.word_max_length:
                    skipped += 1
                    continue
                flag = False
                for g in graphemes:
                    if g not in self.grapheme_set:
                        flag = True
                if flag:
                    continue
                stress_mask = np.zeros(self.word_max_length)
                for stress in secondary:
                    stress_mask[stress] = 2
                for stress in primary:
                    stress_mask[stress] = 1
                x.append(graphemes)
                y.append(stress_mask)
        y = np.array(y)
        logging.debug("Skipped: " + str(skipped))
        return x, y

    def __prepare_data(self, x: List[str], y: np.array = None) -> Tuple[np.array, List[int]]:
        """
        Подготовка данных

        :param x: семплы.
        :param y: ответы.
        :return: очищенные семплы и овтеты.
        """
        x = [[self.grapheme_set.find(ch) for ch in p] for p in x]
        x = sequence.pad_sequences(x, maxlen=self.word_max_length, padding='post', truncating='post')
        if y is not None:
            y = y.reshape((y.shape[0], y.shape[1], 1))
        return x, y

    def __evaluate_wer(self, x: np.array, y: np.array) -> Tuple[float, float]:
        """
        Считаем word error rate - количество слов, в которых была допущена 
        хоть одна ошибка при транскрибировании.

        :param x: данные.
        :param y: ответы
        :return: метрики.
        """
        print("Validation:")
        answer = self.model.predict(x, verbose=0)
        word_errors = 0
        stress_errors = 0
        for i, word in enumerate(answer):
            flag = False
            for j, ch_prob in enumerate(word):
                a = np.argmax(ch_prob)
                if y[i][j] != a:
                    stress_errors += 1
                    flag = True
            if flag:
                word_errors += 1

        wer = float(word_errors) / answer.shape[0]
        all_phonemes = np.ndarray.flatten(y)
        all_phonemes = all_phonemes[all_phonemes != 0]
        per = float(stress_errors) / len(all_phonemes)
        print("WER: " + str(wer))
        print("PER: " + str(per))
        return wer, per



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты к классификатору метра.

import unittest
import jsonpickle
import copy
import logging
import sys

from rupo.main.markup import Markup
from rupo.stress.predictor import CombinedStressPredictor
from rupo.metre.metre_classifier import MetreClassifier, ClassificationResult, StressCorrection
from rupo.settings import RU_STRESS_DEFAULT_MODEL, ZALYZNYAK_DICT, CMU_DICT, RU_GRAPHEME_STRESS_PATH, \
    RU_GRAPHEME_STRESS_TRIE_PATH


class TestMetreClassifier(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stress_predictor = CombinedStressPredictor(
            stress_model_path=RU_STRESS_DEFAULT_MODEL,
            zalyzniak_dict=ZALYZNYAK_DICT,
            cmu_dict=CMU_DICT,
            raw_stress_dict_path=RU_GRAPHEME_STRESS_PATH,
            stress_trie_path=RU_GRAPHEME_STRESS_TRIE_PATH
        )
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    @classmethod
    def tearDownClass(cls):
        del cls.stress_predictor

    def test_classification_result(self):
        result = ClassificationResult(5)
        result.additions["iambos"].append(StressCorrection(0, 0, 0, "", 0))
        self.assertEqual(result, jsonpickle.decode(result.to_json()))

    def test_metre_classifier1(self):
        text = "Горит восток зарёю новой.\n" \
               "Уж на равнине, по холмам\n" \
               "Грохочут пушки. Дым багровый\n" \
               "Кругами всходит к небесам."
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertIsInstance(markup, Markup)
        self.assertIsInstance(result, ClassificationResult)
        self.assertEqual(result.metre, "iambos")

    def test_metre_classifier2(self):
        text = "Буря мглою небо кроет,\n" \
               "Вихри снежные крутя;\n" \
               "То, как зверь, она завоет,\n" \
               "То заплачет, как дитя..."
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertEqual(result.metre, "choreios")

    def test_metre_classifier3(self):
        text = "На стеклах нарастает лед,\n"\
               "Часы твердят: «Не трусь!»\n"\
               "Услышать, что ко мне идет,\n"\
               "И мертвой я боюсь.\n"\
               "Как идола, молю я дверь;\n"\
               "«Не пропускай беду!»\n"\
               "Кто воет за стеной, как зверь,\n"\
               "Кто прячется в саду?"
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertEqual(result.metre, "iambos")

    def test_metre_classifier4(self):
        text = "Вот уж вечер. Роса\n" \
               "Блестит на крапиве.\n"\
               "Я стою у дороги,\n"\
               "Прислонившись к иве.\n"\
               "От луны свет большой\n"\
               "Прямо на нашу крышу.\n"\
               "Где-то песнь соловья\n"\
               "Хорошо и тепло,\n"\
               "Как зимой у печки.\n"\
               "И березы стоят,\n"\
               "Как большие свечки.\n"\
               "И вдали за рекой,\n"\
               "Видно, за опушкой,\n"\
               "Сонный сторож стучит\n"\
               "Мертвой колотушкой."
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertTrue(result.metre == "dolnik3" or result.metre == "dolnik2")

    def test_metre_classifier5(self):
        text = "Глыбу кварца разбили молотом,\n" \
               "И, веселым огнем горя,\n" \
               "Заблестели крупинки золота\n" \
               "В свете тусклого фонаря.\n" \
               "И вокруг собрались откатчики:\n" \
               "Редкий случай, чтоб так, в руде!\n" \
               "И от ламп заплясали зайчики,\n" \
               "Отражаясь в черной воде...\n" \
               "Прислонившись к мокрой стене,\n" \
               "Мы стояли вокруг.\n" \
               "Курили,\n" \
               "Прислонившись к мокрой стене,\n" \
               "И мечтательно говорили\n" \
               "Не о золоте — о весне.\n" \
               "И о том, что скоро, наверно,\n" \
               "На заливе вспотеет лед\n" \
               "И, снега огласив сиреной,\n" \
               "Наконец придет пароход...\n" \
               "Покурили еще немного,\n" \
               "Золотинки в кисет смели\n" \
               "И опять — по своим дорогам,\n" \
               "К вагонеткам своим пошли.\n" \
               "Что нам золото? В дни тяжелые\n" \
               "Я от жадности злой не слеп.\n" \
               "Самородки большие, желтые\n" \
               "Отдавал за табак и хлеб.\n" \
               "Не о золоте были мысли...\n" \
               "В ночь таежную у костра\n" \
               "Есть над чем поразмыслить в жизни,\n" \
               "Кроме\n" \
               "Золота-серебра."
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertTrue(result.metre == "dolnik3" or result.metre == "dolnik2")

    def test_metre_classifier6(self):
        text = "Лючинь печальная читала вечером ручьисто-вкрадчиво,\n" \
               "Так чутко чувствуя журчащий вычурно чужой ей плач,\n" \
               "И, в человечестве чтя нечто вечное, чем чушь Бокаччио,\n" \
               "От чар отчаянья кручинно-скучная, чла час удач."
        markup, result = MetreClassifier.improve_markup(Markup.process_text(text, self.stress_predictor))
        self.assertTrue(result.metre == "iambos")

    def test_improve(self):
        text = "Буря мглою небо кроет,\n" \
               "Вихри снежные крутя;\n" \
               "То, как зверь, она завоет,\n" \
               "То заплачет, как дитя..."
        initial_markup = Markup.process_text(text, self.stress_predictor)
        self.assertEqual(initial_markup.lines[0].words[0].syllables[0].stress, -1)
        markup, result = MetreClassifier.improve_markup(copy.deepcopy(initial_markup))
        self.assertNotEqual(markup.lines[0].words[0].syllables[0].stress, -1)
        self.assertEqual(markup.lines[0].words[0].syllables[1].stress, -1)







import pymorphy2
from russian_tagsets import converters
from rupo.main.tokenizer import Tokenizer, Token


class PyMoprhyRussianMorphology:
    @staticmethod
    def pymorphy_process(input_filename, output_filename):
        """
        Сделать морфоразметку на вход генератору с помощью pymorphy2 и russian-tagsets.
        
        :param input_filename: входной файл - raw текст.
        :param output_filename: выходной файл - разметка.
        """
        morph = pymorphy2.MorphAnalyzer()
        to_ud = converters.converter('opencorpora-int', 'ud14')
        with open(input_filename, "r", encoding="utf-8") as inp:
            with open(output_filename, "w", encoding="utf-8") as out:
                for line in inp:
                    tokens = Tokenizer.tokenize(line)
                    accepted_types = [Token.TokenType.WORD]
                    tokens = [token for token in tokens if token.token_type in accepted_types]
                    for token in tokens:
                        text = token.text.lower()
                        parse = morph.parse(text)[0]
                        lemma = parse.normal_form
                        ud_tag = to_ud(str(parse.tag), text)
                        pos = ud_tag.split()[0]
                        gram = ud_tag.split()[1]
                        out.write("%s\t%s\t%s\t%s\n" % (text, lemma, pos, gram))
                    out.write("\n")

    @staticmethod
    def vocab_process(input_filename, output_filename, word_form_vocabulary, grammeme_vectorizer):
        """
        Сделать морфоразметку по словарю словоформ.
        
        :param input_filename: входной файл - raw текст.
        :param output_filename: выходной файл - разметка.
        :param word_form_vocabulary: слоаврь словоформ.
        :param grammeme_vectorizer: векторы граммем.
        :return: 
        """
        with open(input_filename, "r", encoding="utf-8") as r:
            with open(output_filename, "w", encoding="utf-8") as w:
                for line in r:
                    tokens = Tokenizer.tokenize(line)
                    accepted_types = [Token.TokenType.WORD]
                    tokens = [token for token in tokens if token.token_type in accepted_types]
                    for token in tokens:
                        text = token.text.lower()
                        word_forms = word_form_vocabulary.get_word_forms_by_text(text)
                        if len(word_forms) == 0:
                            print(text)
                        else:
                            indices = [word_form_vocabulary.get_word_form_index(form) for form in word_forms]
                            index = min(indices)
                            word_form = word_form_vocabulary.get_word_form_by_index(index)
                            name = grammeme_vectorizer.get_name_by_index(word_form.gram_vector_index)
                            pos = name.split("#")[0]
                            gram = name.split("#")[1]
                            w.write("%s\t%s\t%s\t%s\n" % (text, word_form.lemma, pos, gram))
                    w.write("\n")



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Конвертер CMU словаря.


class CMUDict:
    aprabet2ipa = {
        "AO": "ɔ",
        "AA": "ɑ",
        "IY": "i",
        "UW": "u",
        "EH": "ɛ",
        "IH": "ɪ",
        "UH": "ʊ",
        "AH": "ʌ",
        "AX": "ə",
        "AE": "æ",
        "EY": "eɪ",
        "AY": "aɪ",
        "OW": "oʊ",
        "AW": "aʊ",
        "OY": "ɔɪ",
        "ER": "ɝ",
        "AXR": "ɚ",
        "P": "p",
        "B": "b",
        "T": "t",
        "D": "d",
        "K": "k",
        "G": "ɡ",
        "CH": "ʦ",
        "JH": "ʤ",
        "F": "f",
        "V": "v",
        "TH": "θ",
        "DH": "ð",
        "S": "s",
        "Z": "z",
        "SH": "ʃ",
        "ZH": "ʒ",
        "HH": "h",
        "M": "m",
        "EM": "m",
        "N": "n",
        "EN": "n",
        "NG": "ŋ",
        "ENG": "ŋ",
        "L": "ɫ",
        "EL": "ɫ",
        "R": "r",
        "DX": "ɾ",
        "NX": "ɾ",
        "Y": "j",
        "W": "w",
        "Q": "ʔ"
    }

    diphtongs = ["EY", "AY", "OW", "AW", "OY"]

    @staticmethod
    def convert_to_g2p_only(source_file, destination_file):
        clean = []
        with open(source_file, 'r', encoding="utf-8", errors="ignore") as f:
            lines = f.readlines()
            for line in lines:
                g = line.split("  ")[0].lower()
                if not ("a" <= g[0] <= "z"):
                    continue
                if "(" in g:
                    continue
                p = line.split("  ")[1].strip()
                phonemes = p.split(" ")
                for i, phoneme in enumerate(phonemes):
                    if not ("A" <= phoneme[-1] <= "Z"):
                        phonemes[i] = phoneme[:-1]
                p = "".join([CMUDict.aprabet2ipa[phoneme] for phoneme in phonemes])
                clean.append((g, p))
        with open(destination_file, 'w', encoding="utf-8") as w:
            for g, p in clean:
                w.write(g+"\t"+p+"\n")

    @staticmethod
    def convert_to_phoneme_stress(source_file, destination_file):
        clean = []
        with open(source_file, 'r', encoding="utf-8", errors="ignore") as f:
            for line in f:
                g = line.split("  ")[0].lower()
                if not ("a" <= g[0] <= "z"):
                    continue
                p = line.split("  ")[1].strip()
                if "(1)" in g:
                    g = g.replace("(1)", "")
                if "(2)" in g:
                    g = g.replace("(2)", "")
                if "(" in g:
                    continue

                phonemes = p.split(" ")
                primary = []
                secondary = []
                diphtongs_count = 0
                for i, phoneme in enumerate(phonemes):
                    if not ("A" <= phoneme[-1] <= "Z"):
                        if int(phoneme[-1]) == 1:
                            primary.append(str(i+diphtongs_count))
                        if int(phoneme[-1]) == 2:
                            secondary.append(str(i+diphtongs_count))
                        phonemes[i] = phoneme[:-1]
                        if phonemes[i] in CMUDict.diphtongs:
                            diphtongs_count += 1
                p = "".join([CMUDict.aprabet2ipa[phoneme] for phoneme in phonemes])
                clean.append((p, primary, secondary))
        with open(destination_file, 'w', encoding="utf-8") as w:
            for p, f, s in clean:
                w.write(p + "\t" + ",".join(f) + "\t" + ",".join(s) + "\n")


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для API библиотеки.

import unittest
import os
import random

from rupo.settings import MARKUP_XML_EXAMPLE, EXAMPLES_DIR, GENERATOR_LSTM_MODEL_PATH, \
    GENERATOR_WORD_FORM_VOCAB_PATH, GENERATOR_VOCAB_PATH, GENERATOR_GRAM_VECTORS, RU_STRESS_DEFAULT_MODEL, \
    RU_G2P_DEFAULT_MODEL, ZALYZNYAK_DICT, RU_GRAPHEME_STRESS_PATH, RU_GRAPHEME_STRESS_TRIE_PATH
from rupo.main.markup import Markup
from rupo.api import Engine


class TestApi(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.engine = Engine(language="ru")
        cls.engine.load(
            stress_model_path=RU_STRESS_DEFAULT_MODEL,
            g2p_model_path=RU_G2P_DEFAULT_MODEL,
            zalyzniak_dict=ZALYZNYAK_DICT,
            raw_stress_dict_path=RU_GRAPHEME_STRESS_PATH,
            stress_trie_path=RU_GRAPHEME_STRESS_TRIE_PATH
        )

    @classmethod
    def tearDownClass(cls):
        del cls.engine

    def test_stress(self):
        self.assertEqual(self.engine.get_stresses("корова"), [3])
        self.assertCountEqual(self.engine.get_stresses("авиамоделирование"), [0, 9])
        self.assertEqual(self.engine.get_stresses("триплекс"), [2])
        self.assertEqual(self.engine.get_stresses("горит"), [3])
        self.assertEqual(self.engine.get_stresses("восток"), [4])
        self.assertEqual(self.engine.get_stresses("зарёю"), [3])
        self.assertEqual(self.engine.get_stresses("новой"), [1])
        self.assertEqual(self.engine.get_stresses("равнине"), [4])
        self.assertEqual(self.engine.get_stresses("холмам"), [4])
        self.assertEqual(self.engine.get_stresses("грохочут"), [4])
        self.assertCountEqual(self.engine.get_stresses("пушки"), [4, 1])
        self.assertEqual(self.engine.get_stresses("багровый"), [4])
        self.assertEqual(self.engine.get_stresses("кругами"), [4])
        self.assertEqual(self.engine.get_stresses("уж"), [0])
        self.assertEqual(self.engine.get_stresses('колесом'), [5])

    def test_get_word_syllables(self):
        self.assertEqual(self.engine.get_word_syllables("корова"), ["ко", "ро", "ва"])

    def test_count_syllables(self):
        self.assertEqual(self.engine.count_syllables("корова"), 3)

    def test_is_rhyme(self):
        self.assertTrue(self.engine.is_rhyme("корова", "здорова"))

    def test_get_markup(self):
        self.assertIsInstance(self.engine.get_markup("корова"), Markup)

    def test_get_improved_markup(self):
        self.assertIsInstance(self.engine.get_improved_markup("корова")[0], Markup)

    def test_classify_metre(self):
        text = "Горит восток зарёю новой.\n" \
               "Уж на равнине, по холмам\n" \
               "Грохочут пушки. Дым багровый\n" \
               "Кругами всходит к небесам."
        self.assertEqual(self.engine.classify_metre(text), "iambos")

    def test_markov_generate_poem(self):
        vocab_dump_file = os.path.join(EXAMPLES_DIR, "vocab.pickle")
        markov_dump_file = os.path.join(EXAMPLES_DIR, "markov.pickle")
        self.assertIsNotNone(
            self.engine.generate_markov_poem(MARKUP_XML_EXAMPLE, markov_dump_file, vocab_dump_file,
                                             rhyme_pattern="a", n_syllables=4, beam_width=20, metre_schema="-+"))
        os.remove(vocab_dump_file)
        os.remove(markov_dump_file)
        self.engine.vocabulary = None
        self.engine.markov = None
        self.engine.markov_generator = None

    def test_lstm_generate_poem(self):
        if os.path.exists(GENERATOR_LSTM_MODEL_PATH) and \
                os.path.exists(GENERATOR_WORD_FORM_VOCAB_PATH) and \
                os.path.exists(GENERATOR_VOCAB_PATH):
            random.seed(42)
            poem = self.engine.generate_poem(GENERATOR_LSTM_MODEL_PATH, GENERATOR_WORD_FORM_VOCAB_PATH,
                                             GENERATOR_GRAM_VECTORS, GENERATOR_VOCAB_PATH, beam_width=30, n_syllables=4,
                                             rhyme_pattern="aa", metre_schema="-+")
            self.assertIsNotNone(poem)

    def test_get_word_rhymes(self):
        vocab_dump_file = os.path.join(EXAMPLES_DIR, "vocab.pickle")
        self.assertEqual(self.engine.get_word_rhymes("глядел", vocab_dump_file, MARKUP_XML_EXAMPLE), ["сидел", "летел"])
        os.remove(vocab_dump_file)



# -*- coding: utf-8 -*-
# Авторы: Анастасьев Даниил, Гусев Илья
# Описание: Модуль рекуррентой сети для генерации языковой модели.

import numpy as np
from typing import List, Tuple
from itertools import islice
from keras.models import model_from_json,  Model, load_model
from keras.layers import Input, Embedding, Dense, LSTM, SpatialDropout1D, Masking, \
    BatchNormalization, Activation, concatenate

from rupo.generate.word_form import WordForm
from rupo.generate.word_form_vocabulary import WordFormVocabulary, SEQ_END_WF
from rupo.generate.corpora_information_loader import CorporaInformationLoader
from rupo.generate.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.model_container import ModelContainer
from rupo.generate.tqdm_open import tqdm_open
from rupo.settings import GENERATOR_LSTM_MODEL_PATH, GENERATOR_WORD_FORM_VOCAB_PATH, GENERATOR_GRAM_VECTORS


class BatchGenerator:
    """
    Генератор наборов примеров для обучения.
    """
    def __init__(self, filenames: List[str], batch_size: int, 
                 embedding_size: int, softmax_size: int, sentence_maxlen: int,
                 word_form_vocabulary: WordFormVocabulary, grammeme_vectorizer: GrammemeVectorizer):
        """
        :param filenames: имена файлов с морфоразметкой.
        :param batch_size: размер набора семплов.
        :param softmax_size: размер выхода softmax-слоя (=размер итогового набора вероятностей)
        :param sentence_maxlen: маскимальная длина куска предложения.
        :param word_form_vocabulary: словарь словофрм.
        :param grammeme_vectorizer: векторизатор граммем.
        """
        self.filenames = filenames  # type: List[str]
        self.batch_size = batch_size  # type: int
        self.embedding_size = embedding_size # type: int
        self.softmax_size = softmax_size  # type: int
        self.sentence_maxlen = sentence_maxlen  # type: int
        self.word_form_vocabulary = word_form_vocabulary  # type: WordFormVocabulary
        self.grammeme_vectorizer = grammeme_vectorizer  # type: GrammemeVectorizer

    def __generate_seqs(self, sentences: List[List[WordForm]]) -> Tuple[List[List[WordForm]], List[WordForm]]:
        """
        Генерация семплов.
        
        :param sentences: куски предложений из словоформ.
        :return: пары (<семпл из словоформ>, следующая за ним словоформа (ответ)).
        """
        seqs, next_words = [], []
        for sentence in sentences:
            # Разворот для генерации справа налево.
            sentence = sentence[::-1]
            for i in range(1, len(sentence)):
                word_form = sentence[i]
                current_part = sentence[max(0, i-self.sentence_maxlen): i]
                if sum([self.word_form_vocabulary.get_lemma_index(x) >= self.embedding_size
                        for x in current_part]) != 0:
                    continue
                # Если следующая словооформа не из предсказываемых, пропускаем её.
                if self.word_form_vocabulary.get_word_form_index(word_form) >= self.softmax_size:
                    continue
                seqs.append(current_part)
                next_words.append(word_form)
        return seqs, next_words

    def __to_tensor(self, sentences: List[List[WordForm]], next_words: List[WordForm]) -> \
            Tuple[np.array, np.array, np.array]:
        """
        Перевод семплов из словоформ в индексы словоформ, поиск грамматических векторов по индексу.
        
        :param sentences: семплы из словоформ.
        :param next_words: следующие за последовательностями из sentences слова.
        :return: индексы словоформ, грамматические векторы, ответы-индексы.
        """
        n_samples = len(sentences)
        max_len = self.sentence_maxlen
        lemmas = np.zeros((n_samples, max_len), dtype=np.int)
        grammemes = np.zeros((n_samples, max_len, self.grammeme_vectorizer.grammemes_count()), dtype=np.int)
        y = np.zeros(n_samples, dtype=np.int)
        for i in range(n_samples):
            sentence = sentences[i]
            next_word = next_words[i]
            lemmas[i, -len(sentence):] = [min(self.word_form_vocabulary.get_lemma_index(x), self.embedding_size)
                                          for x in sentence]
            grammemes[i, -len(sentence):] = [self.grammeme_vectorizer.get_vector_by_index(x.gram_vector_index)
                                             for x in sentence]
            y[i] = min(self.word_form_vocabulary.word_form_indices[next_word], self.softmax_size)
        return lemmas, grammemes, y

    def __iter__(self):
        """
        Получение очередного батча.
        
        :return: индексы словоформ, грамматические векторы, ответы-индексы.
        """
        sentences = [[]]
        for filename in self.filenames:
            with tqdm_open(filename, encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if len(line) == 0:
                        sentences[-1].append(SEQ_END_WF)
                        sentences.append([])
                    else:
                        try:
                            word, lemma, pos, tags = line.split('\t')[:4]
                        except ValueError:
                            pass
                        word, lemma = word.lower(), lemma.lower() + '_' + pos
                        tags = "|".join(sorted(tags.split("|")))
                        gram_vector_index = self.grammeme_vectorizer.name_to_index[pos+"#"+tags]
                        sentences[-1].append(WordForm(lemma, gram_vector_index, word))
                    if len(sentences) >= self.batch_size:
                        sentences, next_words = self.__generate_seqs(sentences)
                        yield self.__to_tensor(sentences, next_words)
                        sentences = [[]]


class LSTMGenerator:
    """
    Языковая модель на основе двухуровневой LSTM RNN.
    """
    def __init__(self, embedding_size: int=30000, external_batch_size: int=10000, nn_batch_size: int=768,
                 sentence_maxlen: int=10, lstm_units=368, embeddings_dimension: int=150, 
                 grammeme_dense_units: Tuple[int]=(35, 15), dense_units: int=256, softmax_size: int=60000,
                 recalculate_softmax=False):
        """
        :param embeddings_size: размер входного слоя (=размер словаря)
        :param softmax_size: размер выхода softmax-слоя (=размер итогового набора вероятностей)
        :param external_batch_size: размер набора семплов для BatchGenerator'а.
        :param nn_batch_size: размер набора семплов для обучения.
        :param sentence_maxlen: маскимальная длина куска предложения.
        """
        self.embedding_size = embedding_size  # type: int
        self.softmax_size = softmax_size  # type: int
        self.external_batch_size = external_batch_size  # type: int
        self.nn_batch_size = nn_batch_size  # type: int
        self.sentence_maxlen = sentence_maxlen  # type: int
        self.word_form_vocabulary = None  # type: WordFormVocabulary
        self.grammeme_vectorizer = None  # type: GrammemeVectorizer
        self.lstm_units = lstm_units  # type: int
        self.embeddings_dimension = embeddings_dimension  # type: int
        self.grammeme_dense_units = grammeme_dense_units  # type: List[int]
        self.dense_units = dense_units  # type: int
        self.model = None  # type: Model
        self.recalculate_softmax = recalculate_softmax

    def prepare(self, filenames: List[str]=list(),
                word_form_vocab_dump_path: str=GENERATOR_WORD_FORM_VOCAB_PATH,
                gram_dump_path: str=GENERATOR_GRAM_VECTORS) -> None:
        """
        Подготовка векторизатора грамматических значений и словаря словоформ по корпусу.
        
        :param filenames: имена файлов с морфоразметкой.
        :param word_form_vocab_dump_path: путь к дампу словаря словоформ.
        :param gram_dump_path: путь к векторам грамматических значений.
        """
        self.grammeme_vectorizer = GrammemeVectorizer(gram_dump_path)
        self.word_form_vocabulary = WordFormVocabulary(word_form_vocab_dump_path)
        if self.grammeme_vectorizer.is_empty() or self.word_form_vocabulary.is_empty():
            loader = CorporaInformationLoader()
            self.word_form_vocabulary, self.grammeme_vectorizer = loader.parse_corpora(filenames)
            self.grammeme_vectorizer.save()
            self.word_form_vocabulary.save()
        if self.recalculate_softmax:
            self.softmax_size = self.word_form_vocabulary.get_softmax_size_by_lemma_size(self.embedding_size)
            print("Recalculated softmax: ", self.softmax_size)

    def load(self, model_filename: str) -> None:
        """
        Загрузка модели.
        
        :param model_filename: файл с моделью.
        """
        self.model = load_model(model_filename)

    def load_with_weights(self, json_filename: str, weights_filename: str) -> None:
        """
        Загрузка модели из json описания и файла с весами.
        
        :param json_filename: json описание.
        :param weights_filename: файл с весам.
        """
        json_string = open(json_filename, 'r', encoding='utf8').readline()
        self.model = model_from_json(json_string)
        self.model.load_weights(weights_filename)

    def build(self):
        """
        Описание модели.
        """
        # Вход лемм
        lemmas = Input(shape=(None,), name='lemmas')
        lemmas_embedding = Embedding(self.embedding_size + 1, self.embeddings_dimension, name='embeddings')(lemmas)
        lemmas_embedding = SpatialDropout1D(.3)(lemmas_embedding)

        # Вход граммем
        grammemes_input = Input(shape=(None, self.grammeme_vectorizer.grammemes_count()), name='grammemes')
        grammemes_layer = Masking(mask_value=0.)(grammemes_input)
        for grammeme_dense_layer_units in self.grammeme_dense_units:
            grammemes_layer = Dense(grammeme_dense_layer_units, activation='relu')(grammemes_layer)

        layer = concatenate([lemmas_embedding, grammemes_layer], name="LSTM_input")
        layer = LSTM(self.lstm_units, dropout=.2, recurrent_dropout=.2, return_sequences=True, name='LSTM_1')(layer)
        layer = LSTM(self.lstm_units, dropout=.2, recurrent_dropout=.2, return_sequences=False, name='LSTM_2')(layer)

        layer = Dense(self.dense_units)(layer)
        layer = BatchNormalization()(layer)
        layer = Activation('relu')(layer)

        output = Dense(self.softmax_size + 1, activation='softmax')(layer)

        self.model = Model(inputs=[lemmas, grammemes_input], outputs=[output])
        self.model.compile(loss='sparse_categorical_crossentropy', optimizer='adam')
        print(self.model.summary())

    @staticmethod
    def __get_validation_data(batch_generator, size):
        """
        Берет первые size батчей из batch_generator для валидационной выборки
        """
        lemmas_list, grammemes_list, y_list = [], [], []
        for lemmas, grammemes, y in islice(batch_generator, size):
            lemmas_list.append(lemmas)
            grammemes_list.append(grammemes)
            y_list.append(y)
        return np.vstack(lemmas_list), np.vstack(grammemes_list), np.hstack(y_list)

    def train(self, filenames: List[str], validation_size: int=5,
              validation_verbosity: int=5, dump_model_freq: int=10) -> None:
        """
        Обучение модели.
        
        :param filenames: имена файлов с морфоразметкой.
        :param validation_size: размер val выборки.
        :param validation_verbosity: каждый validation_verbosity-шаг делается валидация.
        :param dump_model_freq: каждый dump_model_freq-шаг сохраняется модель.
        """
        batch_generator = BatchGenerator(filenames,
                                         batch_size=self.external_batch_size,
                                         embedding_size=self.embedding_size,
                                         softmax_size=self.softmax_size,
                                         sentence_maxlen=self.sentence_maxlen,
                                         word_form_vocabulary=self.word_form_vocabulary,
                                         grammeme_vectorizer=self.grammeme_vectorizer)

        lemmas_val, grammemes_val, y_val = LSTMGenerator.__get_validation_data(batch_generator, validation_size)
        for big_epoch in range(0, 1000):
            print('------------Big Epoch {}------------'.format(big_epoch))
            for epoch, (lemmas, grammemes, y) in enumerate(batch_generator):
                if epoch < validation_size:
                    continue
                self.model.fit([lemmas, grammemes], y, batch_size=self.nn_batch_size, epochs=1, verbose=2)

                if epoch != 0 and epoch % validation_verbosity == 0:
                    print('val loss:',
                          self.model.evaluate([lemmas_val, grammemes_val],
                                              y_val, batch_size=self.nn_batch_size * 2, verbose=0))

                indices = [self.word_form_vocabulary.get_sequence_end_index()]
                for _ in range(10):
                    indices.append(self._sample(self.predict(indices)))
                sentence = [self.word_form_vocabulary.get_word_form_by_index(index) for index in indices]
                print('Sentence', str(big_epoch), str(epoch), end=': ')
                for word in sentence[::-1]:
                    print(word.text, end=' ')
                print()

                if epoch != 0 and epoch % dump_model_freq == 0:
                    self.model.save(GENERATOR_LSTM_MODEL_PATH)

    def predict(self, word_indices: List[int]) -> np.array:
        """
        Предсказание вероятностей следующего слова.
        
        :param word_indices: индексы предыдущих слов.
        :return: проекция языковой модели (вероятности следующего слова).
        """
        if len(word_indices) == 0:
            return np.full(self.softmax_size, 1.0 / self.softmax_size, dtype=np.float)
        cur_sent = [self.word_form_vocabulary.get_word_form_by_index(ind) for ind in word_indices]
        x_lemmas = np.zeros((1, len(cur_sent)))
        x_grammemes = np.zeros((1, len(cur_sent), self.grammeme_vectorizer.grammemes_count()))
        for index, word in enumerate(cur_sent):
            x_lemmas[0, index] = self.word_form_vocabulary.get_lemma_index(word)
            x_grammemes[0, index] = self.grammeme_vectorizer.vectors[word.gram_vector_index]
        prob = self.model.predict([x_lemmas, x_grammemes], verbose=0)[0]
        return prob

    @staticmethod
    def _sample(prob: np.array, temperature: float=1.0) -> int:
        """
        Выбор слова по набору вероятностей с заданной температурой (распределение Больцмана).
        
        :param prob: вероятности.
        :param temperature: температура.
        :return: индекс итогового слова.
        """
        prob = prob[:-1]  # Для исключения неизвестных слов.
        prob = np.log(prob) / temperature
        prob = np.exp(prob) / np.sum(np.exp(prob))
        return np.random.choice(len(prob), p=prob)


class LSTMModelContainer(ModelContainer):
    """
    Контейнер для языковой модели на основе LSTM.
    """
    def __init__(self, model_path=GENERATOR_LSTM_MODEL_PATH,
                 word_form_vocab_dump_path: str=GENERATOR_WORD_FORM_VOCAB_PATH,
                 gram_dump_path: str=GENERATOR_GRAM_VECTORS, cut_context=10):
        self.lstm = LSTMGenerator(embedding_size=5000, recalculate_softmax=True)
        self.lstm.prepare(list(), word_form_vocab_dump_path, gram_dump_path)
        self.lstm.load(model_path)
        self.cut_context = cut_context

    def get_model(self, word_indices: List[int]) -> np.array:
        return self.lstm.predict(word_indices[-self.cut_context:])



import os
import unittest

from rupo.morph.pymorphy import PyMoprhyRussianMorphology
from rupo.settings import EXAMPLES_DIR


class TestMorphology(unittest.TestCase):
    def test_markov(self):
        input_filename = os.path.join(EXAMPLES_DIR, "text.txt")
        output_filename = os.path.join(EXAMPLES_DIR, "morph_markup.txt")
        PyMoprhyRussianMorphology.pymorphy_process(input_filename, output_filename)
        with open(output_filename, "r", encoding="utf-8") as f:
            self.assertEqual(f.readlines()[2], "жизни	жизнь	NOUN	Animacy=Inan|Case=Gen|Gender=Fem|Number=Sing\n")


import re

from rupo.settings import RU_GRAPHEME_SET
from rupo.g2p.phonemes import Phonemes


class WikiDict:
    @staticmethod
    def convert_to_g2p_only(source_file, destination_file):
        with open(source_file, 'r', encoding='utf-8') as r:
            lines = r.readlines()
        with open(destination_file, 'w', encoding='utf-8') as w:
            words = []
            phonetic_words = []
            for line in lines:
                words.append(line.split("\t")[0].strip())
                phonetic_words.append(line.split("\t")[1].replace("'", "").replace("ˌ", "").strip())
            for i, word in enumerate(words):
                w.write(word + "\t" + phonetic_words[i] + "\n")

    @staticmethod
    def convert_to_phoneme_stress(source_file, destination_file):
        with open(source_file, 'r', encoding='utf-8') as r:
            lines = r.readlines()
        with open(destination_file, 'w', encoding='utf-8') as w:
            for line in lines:
                word = line.split("\t")[1].strip()
                clean_word = word.replace("'", "").replace("ˌ", "")
                all_stresses = [i for i, ch in enumerate(word) if ch == "'" or ch == "ˌ"]
                primary = [i for i, ch in enumerate(word) if ch == "'"]
                secondary = [i for i, ch in enumerate(word) if ch == "ˌ"]
                primary_stresses = []
                secondary_stresses = []
                for pos in primary:
                    stress = -1
                    for i, ch in enumerate(word[pos:]):
                        if ch in Phonemes.VOWELS:
                            stress = i + pos - all_stresses.index(pos) - 1
                            break
                    if stress != -1:
                        primary_stresses.append(str(stress))
                for pos in secondary:
                    stress = -1
                    for i, ch in enumerate(word[pos:]):
                        if ch in Phonemes.VOWELS:
                            stress = i + pos - all_stresses.index(pos) - 1
                            break
                    if stress != -1:
                        secondary_stresses.append(str(stress))
                w.write(clean_word + "\t" + ",".join(primary_stresses) + "\t" + ",".join(secondary_stresses) + "\n")

    @staticmethod
    def first_clean_up(filename):
        words = []
        phonetic_words = []
        with open(filename, "r") as f:
            lines = f.readlines()
            print(len(lines))
            for line in lines:
                word = line.split("#")[0]
                word = word.lower()
                phonetic_word = line.split("#")[1]
                if "'" not in phonetic_word and "ˈ" not in phonetic_word:
                    continue
                phonetic_word = phonetic_word.split("/")[0].strip()
                phonetic_word = phonetic_word.split("~")[0].strip()
                phonetic_word = phonetic_word.split(";")[0].strip()
                phonetic_word = phonetic_word.split(",")[0].strip()
                phonetic_word = phonetic_word.replace("ˈ", "'")
                phonetic_word = phonetic_word.replace(":", "ː")
                phonetic_word = re.sub(r"[\s̟̥̻.̞]", "", phonetic_word)
                phonetic_word = re.sub(r"[(⁽][^)⁾]*[)⁾]", "", phonetic_word)
                phonetic_word = Phonemes.clean(phonetic_word)
                wrong_chars = [ch for ch in word if ch not in RU_GRAPHEME_SET]
                if len(wrong_chars) != 0:
                    continue
                if len(word) == 0 or len(phonetic_word) == 0:
                    continue
                if sum([1 for ch in word if ch in "еуаоэяиюёы"]) != \
                        sum([1 for ch in phonetic_word if ch in Phonemes.VOWELS]):
                    continue
                words.append(word)
                phonetic_words.append(phonetic_word)
        print(len(words))
        with open(filename, "w") as f:
            for i, word in enumerate(words):
                f.write(word + "\t" + phonetic_words[i] + "\n")


import vk
from datetime import datetime, timedelta
from rupo.generate.lstm import LSTMModelContainer
from rupo.generate.generator import Generator
from rupo.settings import GENERATOR_VOCAB_PATH
from rupo.main.vocabulary import StressVocabulary

if __name__ == "__main__":
    session = vk.Session(access_token=
                         "30e9c2dee1af80d23e789ea6df52c5b4a56fe83fb618391b0e4e0d086442513c9d8e885e94505078d718d")
    vk_api = vk.API(session)
    lstm = LSTMModelContainer()
    print(lstm.lstm.model.summary())
    # lstm.lstm.word_form_vocabulary.inflate_vocab(GENERATOR_VOCAB_PATH, lstm.lstm.softmax_size+2)
    vocabulary = StressVocabulary(GENERATOR_VOCAB_PATH)
    generator = Generator(lstm, vocabulary, lstm.lstm.word_form_vocabulary)
    current_datetime = datetime.now()

    hours = 12
    metre_schemas = ["+-", "-+"]
    rhyme_schemas = ["aabb", "abab", "aabbcc", "abbacc", "abcabc"]
    syllable_numbers = [6, 8, 10, 12]
    for i in range(200):
        poem = generator.generate_poem(metre_schema=metre_schemas[i % len(metre_schemas)],
                                       rhyme_pattern=rhyme_schemas[i % len(rhyme_schemas)],
                                       n_syllables=syllable_numbers[i % len(syllable_numbers)],
                                       beam_width=1)
        print(poem)
        if poem is not None:
            publish_date = (current_datetime + timedelta(hours=hours) - datetime(1970, 1, 1)) / timedelta(seconds=1)
            vk_api.wall.post(message=poem, owner_id=-147938156, publish_date=publish_date, from_group=1)
            hours += 4


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль для описания разметки по ударениям и слогам.

import json
from typing import List, Set
import xml.etree.ElementTree as etree

from dicttoxml import dicttoxml

from rupo.util.preprocess import get_first_vowel_position
from rupo.util.mixins import CommonMixin
from rupo.main.tokenizer import Tokenizer, Token
from rupo.util.timeit import timeit


class Annotation(CommonMixin):
    """
    Класс аннотации.
    Содержит начальную и конечную позицию в тексте, а также текст аннотации.
    """
    def __init__(self, begin: int, end: int, text: str) -> None:
        self.begin = begin
        self.end = end
        self.text = text


class Syllable(Annotation):
    """
    Разметка слога. Включает в себя аннотацию и номер слога, а также ударение.
    Если ударение падает не на этот слог, -1.
    """
    def __init__(self, begin: int, end: int, number: int, text: str, stress: int=-1) -> None:
        super(Syllable, self).__init__(begin, end, text)
        self.number = number
        self.stress = stress

    def vowel(self) -> int:
        """
        :return: позиция гласной буквы этого слога в слове (с 0).
        """
        return get_first_vowel_position(self.text) + self.begin

    def from_dict(self, d: dict) -> 'Syllable':
        self.__dict__.update(d)
        if "accent" in self.__dict__:
            self.stress = self.__dict__["accent"]
        return self


class Word(Annotation):
    """
    Разметка слова. Включает в себя аннотацию слова и его слоги.
    """
    def __init__(self, begin: int, end: int, text: str, syllables: List[Syllable]) -> None:
        super(Word, self).__init__(begin, end, text)
        self.syllables = syllables

    def count_stresses(self) -> int:
        """
        :return: количество ударений в слове.
        """
        return sum(syllable.stress != -1 for syllable in self.syllables)

    def stress(self) -> int:
        """
        :return: последнее ударение в слове, если нет, то -1.
        """
        stress = -1
        for syllable in self.syllables:
            if syllable.stress != -1:
                stress = syllable.stress
        return stress

    def get_stressed_syllables_numbers(self) -> List[int]:
        """
        :return: номера слогов, на которые падают ударения.
        """
        return [syllable.number for syllable in self.syllables if syllable.stress != -1]

    def get_stresses(self) -> Set[int]:
        """
        :return: все ударения.
        """
        stresses = set()
        for syllable in self.syllables:
            if syllable.stress != -1:
                stresses.add(syllable.stress)
        return stresses

    def set_stresses(self, stresses: List[int]) -> None:
        """
        Задать ударения, все остальные убираются.

        :param stresses: позиции ударения в слове.
        """
        for syllable in self.syllables:
            if syllable.vowel() in stresses:
                syllable.stress = syllable.vowel()
            else:
                syllable.stress = -1

    def get_short(self) -> str:
        """
        :return: слово в форме "текст"+"последнее ударение".
        """
        return self.text.lower() + str(self.stress())

    def from_dict(self, d: dict) -> 'Word':
        self.__dict__.update(d)
        syllables = d["syllables"]  # type: List[dict]
        self.syllables = [Syllable(0, 0, 0, "").from_dict(syllable) for syllable in syllables]
        return self

    def to_stressed_word(self):
        from rupo.stress.word import StressedWord, Stress
        return StressedWord(self.text, set([Stress(pos, Stress.Type.PRIMARY) for pos in self.get_stresses()]))

    def __hash__(self) -> int:
        """
        :return: хеш разметки.
        """
        return hash(self.get_short())


class Line(Annotation):
    """
    Разметка строки. Включает в себя аннотацию строки и её слова.
    """
    def __init__(self, begin: int, end: int, text: str, words: List[Word]) -> None:
        super(Line, self).__init__(begin, end, text)
        self.words = words

    def from_dict(self, d) -> 'Line':
        self.__dict__.update(d)
        words = d["words"]  # type: List[dict]
        self.words = [Word(0, 0, "", []).from_dict(word) for word in words]
        return self

    def count_vowels(self):
        num_vowels = 0
        for word in self.words:
            for syllable in word.syllables:
                if get_first_vowel_position(syllable.text) != -1:
                    num_vowels += 1
        return num_vowels


class Markup(CommonMixin):
    """
    Класс данных для разметки в целом с экспортом/импортом в XML и JSON.
    """
    def __init__(self, text: str=None, lines: List[Line]=None) -> None:
        self.text = text
        self.lines = lines
        self.version = 2

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def from_json(self, st) -> 'Markup':
        d = json.loads(st)
        return self.from_dict(d)

    def from_dict(self, d) -> 'Markup':
        self.__dict__.update(d)
        lines = d["lines"]  # type: List[dict]
        self.lines = [Line(0, 0, "", []).from_dict(line) for line in lines]
        return self

    def to_xml(self) -> str:
        """
        Экспорт в XML.

        :return self: строка в формате XML
        """
        return dicttoxml(self.to_dict(), custom_root='markup', attr_type=False).decode('utf-8').replace("\n", "\\n")

    def from_xml(self, xml: str) -> 'Markup':
        """
        Импорт из XML.

        :param xml: XML-разметка
        :return self: получившийся объект Markup
        """
        root = etree.fromstring(xml)
        if root.find("version") is None or int(root.find("version").text) != self.version:
            raise TypeError("Другая версия разметки")
        lines_node = root.find("lines")
        lines = []
        for line_node in lines_node.findall("item"):
            words_node = line_node.find("words")
            words = []
            for word_node in words_node.findall("item"):
                syllables_node = word_node.find("syllables")
                syllables = []
                for syllable_node in syllables_node.findall("item"):
                    stress_node = syllable_node.find("accent") \
                        if syllable_node.find("accent") is not None \
                        else syllable_node.find("stress")
                    stress = int(stress_node.text)
                    syllables.append(Syllable(int(syllable_node.find("begin").text),
                                              int(syllable_node.find("end").text),
                                              int(syllable_node.find("number").text),
                                              syllable_node.find("text").text,
                                              stress))
                words.append(Word(int(word_node.find("begin").text), int(word_node.find("end").text),
                                  word_node.find("text").text, syllables))
            lines.append(Line(int(line_node.find("begin").text), int(line_node.find("end").text),
                              line_node.find("text").text, words))
        self.text = root.find("text").text.replace("\\n", "\n")
        self.lines = lines
        return self

    def from_raw(self, text: str) -> 'Markup':
        """
        Импорт из сырого текста с ударениями в конце слов

        :param text: текст.
        :return: разметка.
        """

        pos = 0
        lines = []
        for line in text.split("\n"):
            if line == "":
                continue
            line_tokens = []
            for word in line.split(" "):
                i = -1
                ch = word[i]
                stress = ""
                while ch.isdigit() or ch == "-":
                    stress += ch
                    i -= 1
                    ch = word[i]
                line_tokens.append((word[:i+1], int(stress[::-1])))
            words = []
            line_begin = pos
            for pair in line_tokens:
                token = pair[0]
                stress = pair[1]
                from rupo.g2p.graphemes import Graphemes
                syllables = Graphemes.get_syllables(token)
                for j in range(len(syllables)):
                    syllables[j].begin += pos
                    syllables[j].end += pos
                word = Word(pos, pos + len(token), token, syllables)
                word.set_stresses([stress])
                words.append(word)
                pos += len(token) + 1
            lines.append(Line(line_begin, pos, " ".join([pair[0] for pair in line_tokens]), words))
        self.text = "\n".join([line.text for line in lines])
        self.lines = lines
        return self

    @staticmethod
    @timeit
    def process_text(text: str, stress_predictor) -> 'Markup':
        """
        Получение начального варианта разметки по слогам и ударениям.

        :param text: текст для разметки
        :param stress_predictor: предсказатель ударений.
        :return markup: разметка по слогам и ударениям
        """
        from rupo.g2p.graphemes import Graphemes
        begin_line = 0
        lines = []
        words = []
        text_lines = text.split("\n")
        for text_line in text_lines:
            tokens = [token for token in Tokenizer.tokenize(text_line) if token.token_type == Token.TokenType.WORD]
            for token in tokens:
                word = Word(begin_line + token.begin, begin_line + token.end, token.text,
                            Graphemes.get_syllables(token.text))
                # Проставляем ударения.
                stresses = stress_predictor.predict(token.text)
                # Сопоставляем ударения слогам.
                if len(word.syllables) > 1:
                    word.set_stresses(stresses)
                words.append(word)
            end_line = begin_line + len(text_line)
            lines.append(Line(begin_line, end_line, text_line, words))
            words = []
            begin_line = end_line + 1
        return Markup(text, lines)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты словаря.

import os
import unittest

from rupo.main.vocabulary import StressVocabulary
from rupo.settings import EXAMPLES_DIR, MARKUP_XML_EXAMPLE


class TestVocabulary(unittest.TestCase):
    def test_vocabulary(self):
        dump_file = os.path.join(EXAMPLES_DIR, "temp.pickle")
        vocabulary = StressVocabulary(dump_file, MARKUP_XML_EXAMPLE)
        self.assertTrue(os.path.exists(dump_file))
        os.remove(dump_file)
        try:
            self.assertTrue(vocabulary.get_word(0) is not None)
        except IndexError:
            self.assertTrue(False)



# -*- coding: utf-8 -*-
# Авторы: Гусев Илья, Анастасьев Даниил
# Описание: Словарь словоформ.

import pickle
import os
from collections import Counter
from typing import List, Dict, Set
from tqdm import tqdm

from rupo.generate.word_form import WordForm, LemmaCase
from rupo.settings import GENERATOR_WORD_FORM_VOCAB_PATH, GENERATOR_VOCAB_PATH
from rupo.main.markup import Word
from rupo.g2p.graphemes import Graphemes

# Индикатор конца последовательности
SEQ_END = '</s>'
SEQ_END_WF = WordForm(SEQ_END, -1, SEQ_END)


class WordFormVocabulary(object):
    """
    Класс словаря словоформ.
    """
    def __init__(self, dump_filename: str=GENERATOR_WORD_FORM_VOCAB_PATH):
        """
        :param dump_filename: путь к дампу словаря.
        """
        self.dump_filename = dump_filename  # type: str
        self.word_forms = []  # type: List[WordForm]
        # WordForm -> index в self.word_forms
        self.word_form_indices = {}  # type: Dict[WordForm, int]
        # WordForm -> lemma index
        self.lemma_indices = {}  # type: Dict[WordForm, int]
        self.text_to_word_form = None
        if os.path.exists(self.dump_filename):
            self.load()

    def save(self) -> None:
        """
        Сохранение словаря.
        """
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def load(self) -> None:
        """
        Загрузка словаря.
        """
        with open(self.dump_filename, "rb") as f:
            vocab = pickle.load(f)
            self.__dict__.update(vocab.__dict__)

    def init_by_vocabulary(self, lemma_counter: Counter, lemma_to_word_forms: Dict[str, Set[WordForm]],
                           lemma_case: Dict[str, LemmaCase]):
        """
        Строит словарь по предподсчитанным данным
        
        :param lemma_counter: Counter по леммам.
        :param lemma_to_word_forms: Отображение из леммы в список известных словоформ для неё (её парадигму)
        :param lemma_case: Отображение из леммы в тип капитализации, известный для этой леммы
        """
        for i, (lemma, _) in enumerate(tqdm(lemma_counter.most_common(), desc="Init vocabulary")):
            for word_form in lemma_to_word_forms[lemma]:
                word_form.set_case(lemma_case[word_form.lemma])
                self.word_forms.append(word_form)
                self.word_form_indices[word_form] = len(self.word_forms) - 1
                assert self.word_forms[self.word_form_indices[word_form]] == word_form
                self.lemma_indices[word_form] = i + 1  # 0 - зарезервирован для паддинга.
        assert self.lemma_indices[SEQ_END_WF] == 1

    def get_word_form_index(self, word_form: WordForm) -> int:
        return self.word_form_indices[word_form]
            
    def get_word_form_by_index(self, index: int) -> WordForm:
        return self.word_forms[index]

    def get_word_form_index_min(self, word_form: WordForm, size: int) -> int:
        return min(self.get_word_form_index(word_form), size)

    def get_lemma_index_min(self, word_form: WordForm, size: int) -> int:
        return min(self.get_lemma_index(word_form), size)

    def get_lemma_index(self, word_form: WordForm) -> int:
        return self.lemma_indices[word_form]

    def get_sequence_end_index(self) -> int:
        """
        Возвращает индекс словоформы завершающего строку символа.
        """
        assert SEQ_END_WF in self.word_form_indices and self.word_form_indices[SEQ_END_WF] == 0
        return 0

    def get_sequence_end_lemma_index(self) -> int:
        """
        Возвращает индекс леммы завершающего строку символа.
        """
        assert SEQ_END_WF in self.lemma_indices and self.lemma_indices[SEQ_END_WF] == 1
        return 1

    def get_softmax_size_by_lemma_size(self, lemma_size: int):
        assert lemma_size + 1 < len(self.lemma_indices)
        final_lemma = SEQ_END
        for form, index in self.lemma_indices.items():
            if index == lemma_size + 1:
                final_lemma = form.lemma
                break
        for i, form in enumerate(self.word_forms):
            if form.lemma == final_lemma:
                return i
        assert False

    def inflate_vocab(self, dump_path, top_n=None) -> None:
        """
        Получение словаря с ударениями по этому словарю.
        
        :param top_n: сколько первых записей взять?
        :param dump_path: путь, куда сохранить словарь.
        """
        from rupo.main.vocabulary import StressVocabulary
        from rupo.stress.word import StressedWord, Stress
        from rupo.stress.predictor import CombinedStressPredictor
        vocab = StressVocabulary(dump_path)
        stress_predictor = CombinedStressPredictor()
        forms = self.word_forms
        if top_n is not None:
            forms = forms[:top_n]
        for index, word_form in tqdm(enumerate(forms), desc="Accenting words"):
            text = word_form.text
            stresses = [Stress(pos, Stress.Type.PRIMARY) for pos in stress_predictor.predict(text)]
            word = StressedWord(text, set(stresses))
            vocab.add_word(word, index)
        vocab.save()

    def inflate_text_mappings(self):
        self.text_to_word_form = {}
        for i, form in enumerate(self.word_forms):
            self.text_to_word_form[form.text] = i

    def get_word_form_by_text(self, text):
        if self.text_to_word_form is None:
            self.inflate_text_mappings()
        return self.word_forms[self.text_to_word_form[text]]

    def is_empty(self) -> int:
        return len(self.word_forms) == 0



# -*- coding: utf-8 -*-
# Авторы: Анастасьев Даниил
# Описание: Обертка открытия больших файлов в счетчик tqdm

from contextlib import contextmanager
from os.path import getsize, basename
from tqdm import tqdm


@contextmanager
def tqdm_open(filename, encoding='utf8'):
    """
    Открытие файла, обёрнутое в tqdm
    """
    total = getsize(filename)

    def wrapped_line_iterator(fd):
        with tqdm(total=total, unit="B", unit_scale=True, desc=basename(filename), miniters=1) as pb:
            processed_bytes = 0
            for line in fd:
                processed_bytes += len(line)
                if processed_bytes >= 1024 * 1024:
                    pb.update(processed_bytes)
                    processed_bytes = 0
                yield line
            pb.update(processed_bytes)

    with open(filename, encoding=encoding) as fd:
        yield wrapped_line_iterator(fd)


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Набор внешних методов для работы с библиотекой.

from typing import List, Tuple, Dict

from rupo.files.reader import FileType, Reader
from rupo.files.writer import Writer
from rupo.generate.markov import MarkovModelContainer
from rupo.generate.lstm import LSTMModelContainer
from rupo.generate.generator import Generator
from rupo.generate.word_form_vocabulary import WordFormVocabulary
from rupo.main.markup import Markup
from rupo.main.vocabulary import StressVocabulary
from rupo.metre.metre_classifier import MetreClassifier, ClassificationResult
from rupo.rhymes.rhymes import Rhymes
from rupo.stress.predictor import StressPredictor, CombinedStressPredictor
from rupo.g2p.rnn import RNNG2PModel
from rupo.g2p.graphemes import Graphemes
from rupo.settings import RU_G2P_DEFAULT_MODEL, EN_G2P_DEFAULT_MODEL, ZALYZNYAK_DICT, RU_WIKI_DICT, \
    CMU_DICT, RU_GRAPHEME_SET


class Engine:
    def __init__(self, language="ru"):
        self.language = language  # type: str
        self.vocabulary = None  # type: StressVocabulary
        self.markov = None  # type: MarkovModelContainer
        self.markov_generator = None  # type: Generator
        self.lstm_generator = None  # type: Generator
        self.g2p_models = dict()  # type: Dict[str, RNNG2PModel]
        self.stress_predictors = dict()  # type: Dict[str, StressPredictor]

    def load(self, stress_model_path: str=None, g2p_model_path: str=None, raw_stress_dict_path=None,
             stress_trie_path=None, zalyzniak_dict=ZALYZNYAK_DICT):
        self.g2p_models = dict()
        self.stress_predictors = dict()
        self.get_stress_predictor(self.language, stress_model_path, raw_stress_dict_path,
                                  stress_trie_path, zalyzniak_dict)
        self.get_g2p_model(self.language, g2p_model_path)

    def get_vocabulary(self, dump_path: str, markup_path: str) -> StressVocabulary:
        if self.vocabulary is None:
            self.vocabulary = StressVocabulary(dump_path, markup_path)
        return self.vocabulary

    def get_markov(self, dump_path: str, vocab_dump_path: str, markup_path: str, n_grams: int=2,
                   n_poems: int=None) -> MarkovModelContainer:
        if self.markov is None:
            vocab = self.get_vocabulary(vocab_dump_path, markup_path)
            print(vocab.size())
            self.markov = MarkovModelContainer(dump_path, vocab, markup_path, n_grams=n_grams, n_poems=n_poems)
        return self.markov

    def get_markov_generator(self, dump_path: str, vocab_dump_path: str, markup_path: str) -> Generator:
        if self.markov_generator is None:
            self.markov_generator = Generator(self.get_markov(dump_path, vocab_dump_path, markup_path,
                                                              n_grams=2, n_poems=4000),
                                              self.get_vocabulary(vocab_dump_path, markup_path))
        return self.markov_generator

    def get_lstm_generator(self, model_path: str, word_form_vocab_dump_path: str,
                           stress_vocab_dump_path: str, gram_dump_path: str) -> Generator:
        if self.lstm_generator is None:
            lstm = LSTMModelContainer(model_path, word_form_vocab_dump_path, gram_dump_path)
            word_form_vocabulary = WordFormVocabulary(word_form_vocab_dump_path)
            vocabulary = StressVocabulary(stress_vocab_dump_path)
            self.lstm_generator = Generator(lstm, vocabulary, word_form_vocabulary)
        return self.lstm_generator

    def get_stress_predictor(self, language="ru", stress_model_path: str=None, raw_stress_dict_path=None,
                             stress_trie_path=None, zalyzniak_dict=ZALYZNYAK_DICT, cmu_dict=CMU_DICT):
        if self.stress_predictors.get(language) is None:
            self.stress_predictors[language] = CombinedStressPredictor(language, stress_model_path,
                                                                       raw_stress_dict_path, stress_trie_path,
                                                                       zalyzniak_dict, cmu_dict)
        return self.stress_predictors[language]

    def get_g2p_model(self, language="ru", model_path=None):
        if self.g2p_models.get(language) is None:
            self.g2p_models[language] = RNNG2PModel(language=language)
            if language == "ru" and model_path is None:
                model_path = RU_G2P_DEFAULT_MODEL
            elif language == "en" and model_path is None:
                model_path = EN_G2P_DEFAULT_MODEL
            else:
                return None
            self.g2p_models[language].load(model_path)
        return self.g2p_models[language]

    def get_stresses(self, word: str, language: str="ru") -> List[int]:
        """
        :param word: слово.
        :param language: язык.
        :return: ударения слова.
        """
        return self.get_stress_predictor(language).predict(word)

    @staticmethod
    def get_word_syllables(word: str) -> List[str]:
        """
        :param word: слово.
        :return: его слоги.
        """
        return [syllable.text for syllable in Graphemes.get_syllables(word)]

    @staticmethod
    def count_syllables(word: str) -> int:
        """
        :param word: слово.
        :return: количество слогов в нём.
        """
        return len(Graphemes.get_syllables(word))

    def get_markup(self, text: str, language: str="ru") -> Markup:
        """
        :param text: текст.
        :param language: язык.
        :return: его разметка по словарю.
        """
        return Markup.process_text(text, self.get_stress_predictor(language))

    def get_improved_markup(self, text: str, language: str="ru") -> Tuple[Markup, ClassificationResult]:
        """
        :param text: текст.
        :param language: язык.
        :return: его разметка по словарю, классификатору метру и  ML классификатору.
        """
        markup = Markup.process_text(text, self.get_stress_predictor(language))
        return MetreClassifier.improve_markup(markup)

    def classify_metre(self, text: str, language: str="ru") -> str:
        """
        :param text: текст.
        :param language: язык.
        :return: его метр.
        """
        return MetreClassifier.classify_metre(Markup.process_text(text, self.get_stress_predictor(language))).metre

    def generate_markups(self, input_path: str, input_type: FileType, output_path: str, output_type: FileType) -> None:
        """
        Генерация разметок по текстам.

        :param input_path: путь к папке/файлу с текстом.
        :param input_type: тип файлов с текстов.
        :param output_path: путь к файлу с итоговыми разметками.
        :param output_type: тип итогового файла.
        """
        markups = Reader.read_markups(input_path, input_type, False, self.get_stress_predictor())
        writer = Writer(output_type, output_path)
        writer.open()
        for markup in markups:
            writer.write_markup(markup)
        writer.close()

    def is_rhyme(self, word1: str, word2: str) -> bool:
        """
        :param word1: первое слово.
        :param word2: второе слово.
        :return: рифмуются ли слова.
        """
        markup_word1 = self.get_markup(word1).lines[0].words[0]
        markup_word1.set_stresses(self.get_stresses(word1))
        markup_word2 = self.get_markup(word2).lines[0].words[0]
        markup_word2.set_stresses(self.get_stresses(word2))
        return Rhymes.is_rhyme(markup_word1, markup_word2)

    def generate_markov_poem(self, markup_path: str, dump_path: str, vocab_dump_path: str, metre_schema: str="-+",
                             rhyme_pattern: str="abab", n_syllables: int=8, beam_width=5) -> str:
        """
        Сгенерировать стих по данным из разметок.

        :param markup_path: путь к разметкам.
        :param dump_path: путь, куда сохранять модель.
        :param vocab_dump_path: путь, куда сохранять словарь.
        :param metre_schema: схема метра.
        :param rhyme_pattern: схема рифм.
        :param n_syllables: количество слогов в строке.
        :param beam_width: ширина лучевого поиска.
        :return: стих.
        """
        generator = self.get_markov_generator(dump_path, vocab_dump_path, markup_path)
        return generator.generate_poem(metre_schema, rhyme_pattern, n_syllables, beam_width=beam_width)

    def generate_poem(self, model_path: str, word_form_vocab_dump_path: str, gram_dump_path: str,
                      stress_vocab_dump_path: str, metre_schema: str="-+",
                      rhyme_pattern: str="abab", n_syllables: int=8, beam_width: int=5) -> str:
        """
        Сгенерировать стих.

        :param model_path: путь к модели.
        :param word_form_vocab_dump_path: путь к дампу словаря словоформ.
        :param gram_dump_path: путь к векторам грамматических значений.
        :param stress_vocab_dump_path: путь к словарю ударений.
        :param metre_schema: схема метра.
        :param rhyme_pattern: схема рифм.
        :param n_syllables: количество слогов в строке.
        :param beam_width: ширина лучевого поиска.
        :return: стих. None, если генерация не была успешной.
        """
        generator = self.get_lstm_generator(model_path, word_form_vocab_dump_path,
                                            stress_vocab_dump_path, gram_dump_path)
        return generator.generate_poem(metre_schema, rhyme_pattern, n_syllables, beam_width=beam_width)

    def get_word_rhymes(self, word: str, vocab_dump_path: str, markup_path: str=None) -> List[str]:
        """
        Поиск рифмы для данного слова.

        :param word: слово.
        :param vocab_dump_path: путь, куда сохраняется словарь.
        :param markup_path: путь к разметкам.
        :return: список рифм.
        """
        markup_word = self.get_markup(word).lines[0].words[0]
        markup_word.set_stresses(self.get_stresses(word))
        rhymes = []
        vocabulary = self.get_vocabulary(vocab_dump_path, markup_path)
        for i in range(vocabulary.size()):
            if Rhymes.is_rhyme(markup_word, vocabulary.get_word(i)):
                rhymes.append(vocabulary.get_word(i).text.lower())
        return rhymes



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для рекуррентной сети g2p.
#
from keras.layers import LSTM

from rupo.g2p.rnn import RNNG2PModel
from rupo.settings import EN_PHONEME_STRESS_PATH, ACCENT_CURRENT_MODEL_DIR, EN_G2P_DICT_PATH, \
    G2P_CURRENT_MODEL_DIR, RU_PHONEME_STRESS_PATH, RU_G2P_DICT_PATH
from rupo.stress.phoneme_rnn import RNNPhonemeStressModel


def g2p_ru():
    clf = RNNG2PModel(RU_G2P_DICT_PATH, 30, language="ru", rnn=LSTM, units1=512, units2=512, dropout=0.4,
                      batch_size=128, emb_dimension=50)
    clf.build()
    clf.train(G2P_CURRENT_MODEL_DIR, enable_checkpoints=True)


def stress_ru():
    clf = RNNPhonemeStressModel(RU_PHONEME_STRESS_PATH, 19, language="ru", rnn=LSTM)
    clf.build()
    clf.train(ACCENT_CURRENT_MODEL_DIR, enable_checkpoints=True)


def g2p_en():
    clf = RNNG2PModel(EN_G2P_DICT_PATH, 40, language="en", rnn=LSTM, units1=256, dropout=0.5)
    clf.build()
    clf.train(G2P_CURRENT_MODEL_DIR, enable_checkpoints=True)


def stress_en():
    clf = RNNPhonemeStressModel(EN_PHONEME_STRESS_PATH, 25, language="en", rnn=LSTM)
    clf.build()
    clf.train(ACCENT_CURRENT_MODEL_DIR, enable_checkpoints=True)

# if __name__ == "__main__":
    # stress_ru()
    # stress_en()
    # g2p_en()
    # g2p_ru()


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Выравнивание слова и транскрпции.

import pickle
import os
from typing import List, Tuple, Dict
from collections import defaultdict

from rupo.settings import RU_GRAPHEME_SET
from rupo.g2p.phonemes import Phonemes
from rupo.settings import RU_ALIGNER_DEFAULT_PATH, EN_ALIGNER_DEFAULT_PATH, \
    RU_G2P_DICT_PATH, EN_G2P_DICT_PATH, EN_GRAPHEME_SET, RU_WIKI_DICT, CMU_DICT
from rupo.dict.wiki import WikiDict
from rupo.dict.cmu import CMUDict


class Aligner:
    def __init__(self, language="ru", grapheme_set=None, g2p_dict_path=None, dump_path=None,
                 ru_wiki_dict=RU_WIKI_DICT, cmu_dict=CMU_DICT):
        self.grapheme_set = grapheme_set
        self.g2p_dict_path = g2p_dict_path
        self.dump_path = dump_path
        if language == "ru":
            self.__init_language_defaults(RU_GRAPHEME_SET, RU_G2P_DICT_PATH, RU_ALIGNER_DEFAULT_PATH)
            if not os.path.exists(self.g2p_dict_path):
                WikiDict.convert_to_g2p_only(ru_wiki_dict, self.g2p_dict_path)
        elif language == "en":
            self.__init_language_defaults(EN_GRAPHEME_SET, EN_G2P_DICT_PATH, EN_ALIGNER_DEFAULT_PATH)
            if not os.path.exists(self.g2p_dict_path):
                CMUDict.convert_to_g2p_only(cmu_dict, self.g2p_dict_path)
        self.probability_matrix = None
        if os.path.isfile(self.dump_path):
            self.load(self.dump_path)
        else:
            self.train_from_dict()
            self.save(self.dump_path)

    def __init_language_defaults(self, grapheme_set, g2p_dict_path, dump_path):
        if self.grapheme_set is None:
            self.grapheme_set = grapheme_set
        if self.g2p_dict_path is None:
            self.g2p_dict_path = g2p_dict_path
        if self.dump_path is None:
            self.dump_path = dump_path

    def save(self, filename) -> None:
        with open(filename, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def load(self, filename) -> None:
        with open(filename, "rb") as f:
            aligner = pickle.load(f)
            self.__dict__.update(aligner.__dict__)

    def align(self, graphemes: str, phonemes: str):
        """
        Выравнивание графем и фонем.
        
        :param graphemes: графическое слово. 
        :param phonemes: фонетическое слово.
        :return: выровненные слова.
        """
        assert self.probability_matrix is not None
        trace = Aligner.__build_align_matrix(graphemes, phonemes, self.probability_matrix)
        g, p = Aligner.__process_align_trace(trace, graphemes, phonemes)
        return g, p

    def train_from_dict(self):
        with open(self.dict_path, 'r', encoding='utf-8') as r:
            lines = r.readlines()
            pairs = [tuple(line.strip().split("\t")) for line in lines]
            self.train(pairs)

    def train(self, pairs: List[Tuple[str, str]], n_epochs: int=3):
        """
        Обучение EM-алгоритма над словарём пар.
        
        :param pairs: пары графичесих слов и фонетических слов.
        :param n_epochs: количество итерации обучения.
        """
        phoneme_set = "".join(Phonemes.get_all()).replace(" ", "")
        grapheme_set = self.grapheme_set.replace(" ", "")
        # Сначала задаём равномерное распределение.
        probability_matrix = {g: {p: 1.0/len(phoneme_set) for p in phoneme_set} for g in grapheme_set}
        for _ in range(n_epochs):
            g_p_counts = defaultdict(lambda: defaultdict(lambda: 0.0))
            g_counts = defaultdict(lambda: 0.0)
            # E-шаг.
            for graphemes, phonemes in pairs:
                # Считаем динамику с заданной матрицей весов над матрицей из графем и фонем.
                trace = Aligner.__build_align_matrix(graphemes, phonemes, probability_matrix)
                graphemes, phonemes = Aligner.__process_align_trace(trace, graphemes, phonemes)
                # Увеличиваем счётчики, чтобы потом получить апостериорные вероятности.
                for i in range(len(graphemes)):
                    if graphemes[i] != " " and phonemes[i] != " ":
                        g_p_counts[graphemes[i]][phonemes[i]] += 1
                        g_counts[graphemes[i]] += 1
            # M-шаг. Нормализуем вероятности.
            for g, m in probability_matrix.items():
                for p, prob in m.items():
                    probability_matrix[g][p] = g_p_counts[g][p] / g_counts[g] if g_counts[g] != 0 else 0.0
                    # Заплатка, чтобы ʲ не липла к гласным.
                    if p == "ʲ":
                        probability_matrix[g][p] = 0
        self.probability_matrix = probability_matrix

    @staticmethod
    def align_stresses(aligned_g, aligned_p, stresses, is_grapheme=True):
        index = -1
        spaces_count = 0
        word = aligned_g if is_grapheme else aligned_p
        for i, stress in enumerate(stresses):
            for j in range(len(word)):
                if word[j] == " ":
                    spaces_count += 1
                else:
                    index += 1
                    if index == stress:
                        stresses[i] += spaces_count
        return stresses

    @staticmethod
    def __build_align_matrix(first_string: str, second_string: str,
                             probability_matrix: Dict[str, Dict[str, float]], sigma: float=0.0):
        """
        Динамика на матрице g * p.
        
        :param first_string: графемы.
        :param second_string: фонемы
        :param probability_matrix: матрица вероятностей переходов.
        :param sigma: штрафы на пропуски (del).
        :return: путь в матрице, по которому восстаналивается выравнивание.
        """
        f = len(first_string)
        s = len(second_string)
        matrix = [[0 for j in range(s + 1)] for i in range(f + 1)]
        trace = [[0 for j in range(s + 1)] for i in range(f + 1)]

        for i in range(1, f + 1):
            matrix[i][0] = -i * sigma
        for i in range(1, s + 1):
            matrix[0][i] = -i * sigma
        for i in range(1, f + 1):
            for j in range(1, s + 1):
                indel2 = matrix[i - 1][j] - sigma
                indel1 = matrix[i][j - 1] - sigma
                score1 = probability_matrix[first_string[i-1]][second_string[j-1]]
                replace1 = matrix[i - 1][j - 1] + score1
                scores = [indel2, indel1, replace1]
                matrix[i][j] = max(scores)
                trace[i][j] = scores.index(matrix[i][j])
        return trace

    @staticmethod
    def __process_align_trace(trace: List[List[int]], first_string: str, second_string: str):
        """
        Восстановление выравнивания по пути в матрице.
        
        :param trace: путь.
        :param first_string: графемы. 
        :param second_string: фонемы.
        :return: выравненные графемы и фонемы.
        """
        row = len(first_string)
        col = len(second_string)
        insert_indel = lambda word, pos: word[:pos] + ' ' + word[pos:]
        while row != 0 and col != 0:
            if trace[row][col] == 0:
                second_string = insert_indel(second_string, col)
                row -= 1
            elif trace[row][col] == 1:
                first_string = insert_indel(first_string, row)
                col -= 1
            else:
                row -= 1
                col -= 1
        for i in range(row):
            second_string = insert_indel(second_string, 0)
        for i in range(col):
            first_string = insert_indel(first_string, 0)
        return first_string, second_string



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для модуля рифм.

import unittest

from rupo.stress.word import StressedWord, Stress
from rupo.rhymes.rhymes import Rhymes


class TestRhymes(unittest.TestCase):
    def test_rhyme(self):
        self.assertTrue(Rhymes.is_rhyme(StressedWord("тишь", {Stress(1)}),
                                        StressedWord("грустишь", {Stress(5)})))
        self.assertFalse(Rhymes.is_rhyme(StressedWord("наизусть", {Stress(4)}),
                                         StressedWord("сестра", {Stress(5)})))


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты к компилятору выражений.

import unittest

from rupo.metre.pattern_analyzer import PatternAnalyzer


class TestPatternAnalyzer(unittest.TestCase):
    def test_pattern_analyzer(self):
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "uuu"), ('sss', 0, 3, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "uus"), ('sss', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "usu"), ('sss', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "uss"), ('sss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "suu"), ('sss', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "sus"), ('sss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "ssu"), ('sss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)*", "sss"), ('sss', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(sus)*(u)?", 'suu'), ('sus', 0, 1, False))

        self.assertEqual(PatternAnalyzer.count_errors("((sus)*u)*s", 'susss'), ('susus', 0, 1, False))

        self.assertEqual(PatternAnalyzer.count_errors("(s((s)*u)*)*", 'susss'), ('susss', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s((s)*u)*)*", 'usss'), ('ssss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s((s)*u)*)*", 'suuu'), ('suuu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s((s)*u)*)*", 'suuusuuus'), ('suuusuuus', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(sss((sus)*uss)*)*", 'ssssussususs'), ('ssssussususs', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(sss((sus)*uss)*)*", 'ssssuuuss'), ('ssssususs', 0, 1, False))

        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "uuuu"), ('susu', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "uuus"), ('ssus', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "uusu"), ('susu', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "uuss"), ('suss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "usuu"), ('sssu', 0, 2, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "usus"), ('ssus', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "ussu"), ('sssu', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "usss"), ('ssss', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "suuu"), ('susu', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "suus"), ('ssus', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "susu"), ('susu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "suss"), ('suss', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "ssuu"), ('sssu', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", 'ssus'), ('ssus', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", 'sssu'), ('sssu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("((s)(u)?)*", "ssss"), ('ssss', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)?(S)?", "su"), ('su', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)?(S)?", "ss"), ('su', 0, 1, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)?(S)?", "uS"), ('uS', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)?(S)?", "sS"), ('sS', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)(s)*", "u"), ('u', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)(s)*", "su"), ('su', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)(s)*", "us"), ('us', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)(s)*", "sus"), ('sus', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(s)?(u)(s)*", "uss"), ('uss', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(us)*(uS)(U)?(U)?", "usuS"), ('usuS', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(us)*(uS)(U)?(U)?", "uSUU"), ('uSUU', 0, 0, False))

        self.assertEqual(PatternAnalyzer.count_errors("(su(u)?)*", "su"), ('su', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(su(u)?)*", "suu"), ('suu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(su(u)?)*", "susu"), ('susu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(su(u)?)*", "suusuu"), ('suusuu', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(su(u)?)*", "ssussu"), ('suusuu', 0, 2, False))

        self.assertEqual(PatternAnalyzer.count_errors("(u)?(u)?((s)(u)?(u)?)*(S)(U)?(U)?", "sssuSU"), ('sssuSU', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(u)?(u)?((s)(u)?(u)?)*(S)(U)?(U)?", "ussuSU"), ('ussuSU', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(u)?(u)?((s)(u)?(u)?)*(S)(U)?(U)?", "susuuSU"), ('susuuSU', 0, 0, False))
        self.assertEqual(PatternAnalyzer.count_errors("(u)?(u)?((s)(u)?(u)?)*(S)(U)?(U)?", "uusuuSU"), ('uusuuSU', 0, 0, False))



from typing import List

from rupo.main.markup import Syllable
from rupo.util.preprocess import VOWELS, CLOSED_SYLLABLE_CHARS, get_first_vowel_position


class Graphemes:
    @staticmethod
    def get_syllables(word: str) -> List[Syllable]:
        """
        Разделение слова на слоги.

        :param word: слово для разбивки на слоги.
        :return syllables: массив слогов слова.
        """
        syllables = []
        begin = 0
        number = 0

        # В случае наличия дефиса разбиваем слова на подслова, находим слоги в них, объединяем.
        if "-" in word:
            word_parts = word.split("-")
            word_syllables = []
            last_part_end = 0
            for part in word_parts:
                part_syllables = Graphemes.get_syllables(part)
                if len(part_syllables) == 0:
                    continue
                for i in range(len(part_syllables)):
                    part_syllables[i].begin += last_part_end
                    part_syllables[i].end += last_part_end
                    part_syllables[i].number += len(word_syllables)
                word_syllables += part_syllables
                last_part_end = part_syllables[-1].end + 1
            return word_syllables

        # Для слов или подслов, в которых нет дефиса.
        for i, ch in enumerate(word):
            if ch not in VOWELS:
                continue
            if i + 1 < len(word) - 1 and word[i + 1] in CLOSED_SYLLABLE_CHARS:
                if i + 2 < len(word) - 1 and word[i + 2] in "ьЬ":
                    # Если после сонорного согласного идёт мягкий знак, заканчиваем на нём. ("бань-ка")
                    end = i + 3
                elif i + 2 < len(word) - 1 and word[i + 2] not in VOWELS and \
                        (word[i + 2] not in CLOSED_SYLLABLE_CHARS or word[i + 1] == "й"):
                    # Если после сонорного согласного не идёт гласная или другой сонорный согласный,
                    # слог закрывается на этом согласном. ("май-ка")
                    end = i + 2
                else:
                    # Несмотря на наличие закрывающего согласного, заканчиваем на гласной.
                    # ("со-ло", "да-нный", "пол-ный")
                    end = i + 1
            else:
                # Если после гласной идёт не закрывающая согласная, заканчиваем на гласной. ("ко-гда")
                end = i + 1
            syllables.append(Syllable(begin, end, number, word[begin:end]))
            number += 1
            begin = end
        if get_first_vowel_position(word) != -1:
            # Добиваем последний слог до конца слова.
            syllables[-1] = Syllable(syllables[-1].begin, len(word), syllables[-1].number,
                                     word[syllables[-1].begin:len(word)])
        return syllables


# # -*- coding: utf-8 -*-
# # Автор: Гусев Илья
# # Описание: Тесты генератора.

import os
import pickle
from rupo.settings import DATA_DIR, GENERATOR_LSTM_MODEL_PATH
from rupo.generate.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.word_form_vocabulary import WordFormVocabulary
from rupo.generate.lstm import LSTMGenerator

# if __name__ == "__main__":
#     # filename1 = "/media/data/PoetryMorph.txt"
#     # filename2 = "/media/data/ru-ud-train.txt"
#     dir_name = "/media/data/Datasets/Morpho/clean"
#
#     # vectorizer = GrammemeVectorizer()
#     # vectorizer.collect_grammemes(filename)
#     # print(vectorizer.get_ordered_grammemes())
#     # vectorizer.collect_possible_vectors(filename)
#     # print(vectorizer.vectors)
#     #
#     # vocab = WordFormVocabulary()
#     # vocab.load_from_corpus(filename, grammeme_vectorizer=vectorizer)
#     # print(vocab.word_forms)
#
#     lstm = LSTMGenerator(nn_batch_size=256, embedding_size=5000, recalculate_softmax=True)
#     lstm.prepare([os.path.join(dir_name, filename) for filename in os.listdir(dir_name)])
#     # lstm.build()
#     lstm.load(GENERATOR_LSTM_MODEL_PATH)
#     print(lstm.model.summary())
#     lstm.train([os.path.join(dir_name, filename) for filename in os.listdir(dir_name)], validation_size=1,
#                validation_verbosity=5, dump_model_freq=5)


from setuptools import find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install


class PostDevelopCommand(develop):
    def run(self):
        develop.run(self)


class PostInstallCommand(install):
    def run(self):
        install.run(self)


setup(
    name='rupo',
    packages=find_packages(),
    version='0.2.5',
    description='RuPo: library for russian poetry analysis and generation',
    author='Ilya Gusev',
    author_email='phoenixilya@gmail.com',
    url='https://github.com/IlyaGusev/rupo',
    download_url='https://github.com/IlyaGusev/rupo/archive/0.2.5.tar.gz',
    keywords=['poetry', 'nlp', 'russian'],
    package_data={
        'rupo': ['data/examples/*', 'data/hyphen-tokens.txt']
    },
    install_requires=[
        'dicttoxml>=1.7.4',
        'pygtrie>=2.2',
        'numpy>=1.11.3',
        'scipy>=0.18.1',
        'scikit-learn>=0.18.1',
        'jsonpickle>=0.9.4',
        'tensorflow>=1.1.0',
        'keras==2.0.5',
        'pymorphy2>=0.8',
        'h5py>=2.7.0',
        'russian-tagsets==0.6',
        'tqdm>=4.14.0'
    ],
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',

        'Topic :: Text Processing :: Linguistic',

        'License :: OSI Approved :: Apache Software License',

        'Natural Language :: Russian',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)